#!/usr/bin/python
# Copyright (c) 2024 Felix Fontein <felix@fontein.de>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

DOCUMENTATION = """
module: acme_certificate_order_validate
author: Felix Fontein (@felixfontein)
version_added: 2.24.0
short_description: Validate authorizations of an ACME v2 order
description:
  - Validates pending authorizations of an ACME v2 order.
    This is the second to last step of obtaining a new certificate with the
    L(ACME protocol,https://tools.ietf.org/html/rfc8555) from a Certificate
    Authority such as L(Let's Encrypt,https://letsencrypt.org/).
    This module does not support ACME v1, the original version of the ACME protocol before standardization.
  - This module needs to be used in conjunction with the
    M(community.crypto.acme_certificate_order_create) and
    M(community.crypto.acme_certificate_order_finalize) modules.
seealso:
  - module: community.crypto.acme_certificate_order_create
    description: Create an ACME order.
  - module: community.crypto.acme_certificate_order_finalize
    description: Finalize an ACME order after satisfying the challenges.
  - module: community.crypto.acme_certificate_order_info
    description: Obtain information for an ACME order.
  - name: The Let's Encrypt documentation
    description: Documentation for the Let's Encrypt Certification Authority.
                 Provides useful information for example on rate limits.
    link: https://letsencrypt.org/docs/
  - name: Automatic Certificate Management Environment (ACME)
    description: The specification of the ACME protocol (RFC 8555).
    link: https://tools.ietf.org/html/rfc8555
  - name: ACME TLS ALPN Challenge Extension
    description: The specification of the V(tls-alpn-01) challenge (RFC 8737).
    link: https://www.rfc-editor.org/rfc/rfc8737.html
  - module: community.crypto.acme_challenge_cert_helper
    description: Helps preparing V(tls-alpn-01) challenges.
  - module: community.crypto.acme_inspect
    description: Allows to debug problems.
  - module: community.crypto.acme_certificate_deactivate_authz
    description: Allows to deactivate (invalidate) ACME v2 orders.
extends_documentation_fragment:
  - community.crypto._acme.basic
  - community.crypto._acme.account
  - community.crypto._attributes
  - community.crypto._attributes.actiongroup_acme
  - community.crypto._attributes.files
attributes:
  check_mode:
    support: none
  diff_mode:
    support: none
  safe_file_operations:
    support: full
  idempotent:
    support: full
options:
  challenge:
    description:
      - The challenge to be performed for every pending authorization.
      - Must be provided if there is at least one pending authorization.
      - In case of authorization reuse, or in case of CAs which use External Account Binding
        and other means of validating certificate assurance, it might not be necessary
        to provide this option.
      - Support for V(dns-account-01) and V(dns-persist-01) has been added in community.crypto 3.2.0.
    type: str
    choices:
      - http-01
      - dns-01
      - dns-account-01
      - dns-persist-01
      - tls-alpn-01
  order_uri:
    description:
      - The order URI provided by RV(community.crypto.acme_certificate_order_create#module:order_uri).
    type: str
    required: true
  deactivate_authzs:
    description:
      - "Deactivate authentication objects (authz) in case an error happens."
      - "Authentication objects are bound to an account key and remain valid
         for a certain amount of time, and can be used to issue certificates
         without having to re-authenticate the domain. This can be a security
         concern."
    type: bool
    default: true
"""

EXAMPLES = r"""
---
### Example with HTTP-01 challenge ###

- name: Create a challenge for sample.com using a account key from a variable
  community.crypto.acme_certificate_order_create:
    account_key_content: "{{ account_private_key }}"
    csr: /etc/pki/cert/csr/sample.com.csr
  register: sample_com_challenge

# Alternative first step:
- name: Create a challenge for sample.com using a account key from Hashi Vault
  community.crypto.acme_certificate_order_create:
    account_key_content: >-
      {{ lookup('community.hashi_vault.hashi_vault', 'secret=secret/account_private_key:value') }}
    csr: /etc/pki/cert/csr/sample.com.csr
  register: sample_com_challenge

# Alternative first step:
- name: Create a challenge for sample.com using a account key file
  community.crypto.acme_certificate_order_create:
    account_key_src: /etc/pki/cert/private/account.key
    csr_content: "{{ lookup('file', '/etc/pki/cert/csr/sample.com.csr') }}"
  register: sample_com_challenge

# Perform the necessary steps to fulfill the challenge. For example:
#
# - name: Copy http-01 challenges
#   ansible.builtin.copy:
#     dest: /var/www/{{ item.identifier }}/{{ item.challenges['http-01'].resource }}
#     content: "{{ item.challenges['http-01'].resource_value }}"
#   loop: "{{ sample_com_challenge.challenge_data }}"
#   when: "'http-01' in item.challenges"

- name: Let the challenge be validated
  community.crypto.acme_certificate_order_validate:
    account_key_src: /etc/pki/cert/private/account.key
    order_uri: "{{ sample_com_challenge.order_uri }}"
    challenge: http-01

- name: Retrieve the cert and intermediate certificate
  community.crypto.acme_certificate_order_finalize:
    account_key_src: /etc/pki/cert/private/account.key
    csr: /etc/pki/cert/csr/sample.com.csr
    order_uri: "{{ sample_com_challenge.order_uri }}"
    cert_dest: /etc/httpd/ssl/sample.com.crt
    fullchain_dest: /etc/httpd/ssl/sample.com-fullchain.crt
    chain_dest: /etc/httpd/ssl/sample.com-intermediate.crt

---
### Example with DNS challenge against production ACME server ###

- name: Create a challenge for sample.com using a account key file.
  community.crypto.acme_certificate_order_create:
    acme_directory: https://acme-v01.api.letsencrypt.org/directory
    account_key_src: /etc/pki/cert/private/account.key
    csr: /etc/pki/cert/csr/sample.com.csr
  register: sample_com_challenge

# Perform the necessary steps to fulfill the challenge. For example:
#
# - name: Create DNS records for dns-01 challenges
#   community.aws.route53:
#     zone: sample.com
#     record: "{{ item.key }}"
#     type: TXT
#     ttl: 60
#     state: present
#     wait: true
#     # Note: item.value is a list of TXT entries, and route53
#     # requires every entry to be enclosed in quotes
#     value: "{{ item.value | map('community.dns.quote_txt', always_quote=true) | list }}"
#   loop: "{{ sample_com_challenge.challenge_data_dns | dict2items }}"

- name: Let the challenge be validated
  community.crypto.acme_certificate_order_validate:
    acme_directory: https://acme-v01.api.letsencrypt.org/directory
    account_key_src: /etc/pki/cert/private/account.key
    order_uri: "{{ sample_com_challenge.order_uri }}"
    challenge: dns-01

- name: Retrieve the cert and intermediate certificate
  community.crypto.acme_certificate_order_finalize:
    acme_directory: https://acme-v01.api.letsencrypt.org/directory
    account_key_src: /etc/pki/cert/private/account.key
    csr: /etc/pki/cert/csr/sample.com.csr
    order_uri: "{{ sample_com_challenge.order_uri }}"
    cert_dest: /etc/httpd/ssl/sample.com.crt
    fullchain_dest: /etc/httpd/ssl/sample.com-fullchain.crt
    chain_dest: /etc/httpd/ssl/sample.com-intermediate.crt
"""

RETURN = """
account_uri:
  description: ACME account URI.
  returned: success
  type: str
validating_challenges:
  description: List of challenges whose validation was triggered.
  returned: success
  type: list
  elements: dict
  contains:
    identifier:
      description:
        - The identifier the challenge is for.
      type: str
      returned: always
    identifier_type:
      description:
        - The identifier's type for the challenge.
      type: str
      returned: always
      choices:
        - dns
        - ip
    authz_url:
      description:
        - The URL of the authorization object for this challenge.
      type: str
      returned: always
    challenge_type:
      description:
        - The challenge's type.
      type: str
      returned: always
      choices:
        - http-01
        - dns-01
        - tls-alpn-01
    challenge_url:
      description:
        - The URL of the challenge object.
      type: str
      returned: always
"""

import typing as t

from ansible_collections.community.crypto.plugins.module_utils._acme.acme import (
    create_backend,
    create_default_argspec,
)
from ansible_collections.community.crypto.plugins.module_utils._acme.certificate import (
    ACMECertificateClient,
)
from ansible_collections.community.crypto.plugins.module_utils._acme.errors import (
    ModuleFailException,
)

if t.TYPE_CHECKING:
    from ansible_collections.community.crypto.plugins.module_utils._acme.challenges import (  # pragma: no cover
        Authorization,
    )


def main() -> t.NoReturn:
    argument_spec = create_default_argspec(with_certificate=False)
    argument_spec.update_argspec(
        order_uri={"type": "str", "required": True},
        challenge={
            "type": "str",
            "choices": [
                "http-01",
                "dns-01",
                "dns-account-01",
                "dns-persist-01",
                "tls-alpn-01",
            ],
        },
        deactivate_authzs={"type": "bool", "default": True},
    )
    module = argument_spec.create_ansible_module()

    backend = create_backend(module, needs_acme_v2=False)

    try:
        client = ACMECertificateClient(module=module, backend=backend)
        done = False
        order = None
        try:
            # Step 1: load order
            order = client.load_order()
            client.check_that_authorizations_can_be_used(order)

            # Step 2: find all pending authorizations
            pending_authzs = client.collect_pending_authzs(order)

            # Step 3: figure out challenges to use
            challenges = {}
            for authz in pending_authzs:
                challenges[authz.combined_identifier] = module.params["challenge"]

            missing_challenge_authzs = [k for k, v in challenges.items() if v is None]
            if missing_challenge_authzs:
                missing_challenge_authzs_str = ", ".join(
                    sorted(missing_challenge_authzs)
                )
                raise ModuleFailException(
                    "The challenge parameter must be supplied if there are pending authorizations."
                    f" The following authorizations are pending: {missing_challenge_authzs_str}"
                )

            bad_challenge_authzs = [
                authz.combined_identifier
                for authz in pending_authzs
                if authz.find_challenge(
                    challenge_type=challenges[authz.combined_identifier]
                )
                is None
            ]
            if bad_challenge_authzs:
                authz_challenges_pairs = ", ".join(
                    sorted(
                        f"{authz} with {challenges[authz]}"
                        for authz in bad_challenge_authzs
                    )
                )
                raise ModuleFailException(
                    f"The following authorizations do not support the selected challenges: {authz_challenges_pairs}"
                )

            def is_pending(authz: Authorization) -> bool:
                challenge_name = challenges[authz.combined_identifier]
                challenge_obj = authz.find_challenge(challenge_type=challenge_name)
                return challenge_obj is not None and challenge_obj.status == "pending"

            really_pending_authzs = [
                authz for authz in pending_authzs if is_pending(authz)
            ]

            # Step 4: validate pending authorizations
            authzs_with_challenges_to_wait_for = client.call_validate(
                really_pending_authzs,
                get_challenge=lambda authz: challenges[authz.combined_identifier],
                wait=False,
            )

            done = True
        finally:
            if order and module.params["deactivate_authzs"] and not done:
                client.deactivate_authzs(order)
        module.exit_json(
            changed=len(authzs_with_challenges_to_wait_for) > 0,
            account_uri=client.client.account_uri,
            validating_challenges=[
                {
                    "identifier": authz.identifier,
                    "identifier_type": authz.identifier_type,
                    "authz_url": authz.url,
                    "challenge_type": challenge_type,
                    "challenge_url": challenge.url if challenge else None,
                }
                for authz, challenge_type, challenge in authzs_with_challenges_to_wait_for
            ],
        )
    except ModuleFailException as e:
        e.do_fail(module=module)


if __name__ == "__main__":
    main()



# Copyright (c) 2013, Bradley Young <young.bradley@gmail.com>
# Copyright (c) 2017 Ansible Project
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later
from __future__ import annotations

DOCUMENTATION = r"""
author: Unknown (!UNKNOWN)
name: cartesian
short_description: Returns the cartesian product of lists
description:
  - Takes the input lists and returns a list that represents the product of the input lists.
  - It is clearer with an example, it turns [1, 2, 3], [a, b] into [1, a], [1, b], [2, a], [2, b], [3, a], [3, b].
  - You can see the exact syntax in the examples section.
options:
  _terms:
    description:
      - A set of lists.
    type: list
    elements: list
    required: true
"""

EXAMPLES = r"""
- name: Example of the change in the description
  ansible.builtin.debug:
    msg: "{{ lookup('community.general.cartesian', [1,2,3], [a, b])}}"

- name: loops over the cartesian product of the supplied lists
  ansible.builtin.debug:
    msg: "{{item}}"
  with_community.general.cartesian:
    - "{{list1}}"
    - "{{list2}}"
    - [1, 2, 3, 4, 5, 6]
"""

RETURN = r"""
_list:
  description:
    - List of lists composed of elements of the input lists.
  type: list
  elements: list
"""

from itertools import product

from ansible.errors import AnsibleError
from ansible.plugins.lookup import LookupBase
from ansible.utils.listify import listify_lookup_plugin_terms

from ansible_collections.community.general.plugins.plugin_utils._lookup import check_for_wrong_terms


class LookupModule(LookupBase):
    """
    Create the cartesian product of lists
    """

    def _lookup_variables(self, terms):
        """
        Turn this:
            terms == ["1,2,3", "a,b"]
        into this:
            terms == [[1,2,3], [a, b]]
        """
        results = []
        for x in terms:
            results.append(listify_lookup_plugin_terms(x, templar=self._templar))
        return results

    def run(self, terms, variables=None, **kwargs):
        self.set_options(var_options=variables, direct=kwargs)
        check_for_wrong_terms(self, direct=kwargs)

        terms = self._lookup_variables(terms)

        my_list = terms[:]
        if len(my_list) == 0:
            raise AnsibleError("with_cartesian requires at least one element in each list")

        return [self._flatten(x) for x in product(*my_list)]



#!/usr/bin/python
# Copyright (c) 2024 Felix Fontein <felix@fontein.de>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

DOCUMENTATION = """
module: acme_certificate_order_finalize
author: Felix Fontein (@felixfontein)
version_added: 2.24.0
short_description: Finalize an ACME v2 order
description:
  - Finalizes an ACME v2 order and obtains the certificate and certificate chains.
    This is the final step of obtaining a new certificate with the
    L(ACME protocol,https://tools.ietf.org/html/rfc8555) from a Certificate
    Authority such as L(Let's Encrypt,https://letsencrypt.org/).
    This module does not support ACME v1, the original version of the ACME protocol before standardization.
  - This module needs to be used in conjunction with the
    M(community.crypto.acme_certificate_order_create) and.
    M(community.crypto.acme_certificate_order_validate) modules.
seealso:
  - module: community.crypto.acme_certificate_order_create
    description: Create an ACME order.
  - module: community.crypto.acme_certificate_order_validate
    description: Validate pending authorizations of an ACME order.
  - module: community.crypto.acme_certificate_order_info
    description: Obtain information for an ACME order.
  - name: The Let's Encrypt documentation
    description: Documentation for the Let's Encrypt Certification Authority.
                 Provides useful information for example on rate limits.
    link: https://letsencrypt.org/docs/
  - name: Automatic Certificate Management Environment (ACME)
    description: The specification of the ACME protocol (RFC 8555).
    link: https://tools.ietf.org/html/rfc8555
  - module: community.crypto.certificate_complete_chain
    description: Allows to find the root certificate for the returned fullchain.
  - module: community.crypto.acme_certificate_revoke
    description: Allows to revoke certificates.
  - module: community.crypto.acme_inspect
    description: Allows to debug problems.
  - module: community.crypto.acme_certificate_deactivate_authz
    description: Allows to deactivate (invalidate) ACME v2 orders.
extends_documentation_fragment:
  - community.crypto._acme.basic
  - community.crypto._acme.account
  - community.crypto._acme.certificate
  - community.crypto._attributes
  - community.crypto._attributes.actiongroup_acme
  - community.crypto._attributes.files
attributes:
  check_mode:
    support: none
  diff_mode:
    support: none
  safe_file_operations:
    support: full
  idempotent:
    support: full
options:
  order_uri:
    description:
      - The order URI provided by RV(community.crypto.acme_certificate_order_create#module:order_uri).
    type: str
    required: true
  cert_dest:
    description:
      - "The destination file for the certificate."
    type: path
  fullchain_dest:
    description:
      - "The destination file for the full chain (that is, a certificate followed
         by chain of intermediate certificates)."
    type: path
  chain_dest:
    description:
      - If specified, the intermediate certificate will be written to this file.
    type: path
  deactivate_authzs:
    description:
      - "Deactivate authentication objects (authz) after issuing a certificate,
         or when issuing the certificate failed."
      - V(never) never deactivates them.
      - V(always) always deactivates them in cases of errors or when the certificate was issued.
      - V(on_error) only deactivates them in case of errors.
      - V(on_success) only deactivates them in case the certificate was successfully issued.
      - "Authentication objects are bound to an account key and remain valid
         for a certain amount of time, and can be used to issue certificates
         without having to re-authenticate the domain. This can be a security
         concern."
    type: str
    choices:
      - never
      - on_error
      - on_success
      - always
    default: always
  retrieve_all_alternates:
    description:
      - "When set to V(true), will retrieve all alternate trust chains offered by the ACME CA.
         These will not be written to disk, but will be returned together with the main
         chain as RV(all_chains). See the documentation for the RV(all_chains) return
         value for details."
    type: bool
    default: false
  select_chain:
    description:
      - "Allows to specify criteria by which an (alternate) trust chain can be selected."
      - "The list of criteria will be processed one by one until a chain is found
         matching a criterium. If such a chain is found, it will be used by the
         module instead of the default chain."
      - "If a criterium matches multiple chains, the first one matching will be
         returned. The order is determined by the ordering of the C(Link) headers
         returned by the ACME server and might not be deterministic."
      - "Every criterium can consist of multiple different conditions, like O(select_chain[].issuer)
         and O(select_chain[].subject). For the criterium to match a chain, all conditions must apply
         to the same certificate in the chain."
      - "This option can only be used with the C(cryptography) backend."
    type: list
    elements: dict
    suboptions:
      test_certificates:
        description:
          - "Determines which certificates in the chain will be tested."
          - "V(all) tests all certificates in the chain (excluding the leaf, which is
             identical in all chains)."
          - "V(first) only tests the first certificate in the chain, that is the one which
             signed the leaf."
          - "V(last) only tests the last certificate in the chain, that is the one furthest
             away from the leaf. Its issuer is the root certificate of this chain."
        type: str
        default: all
        choices: [first, last, all]
      issuer:
        description:
          - "Allows to specify parts of the issuer of a certificate in the chain must
             have to be selected."
          - "If O(select_chain[].issuer) is empty, any certificate will match."
          - 'An example value would be V({"commonName": "My Preferred CA Root"}).'
        type: dict
      subject:
        description:
          - "Allows to specify parts of the subject of a certificate in the chain must
             have to be selected."
          - "If O(select_chain[].subject) is empty, any certificate will match."
          - 'An example value would be V({"CN": "My Preferred CA Intermediate"})'
        type: dict
      subject_key_identifier:
        description:
          - "Checks for the SubjectKeyIdentifier extension. This is an identifier based
             on the private key of the intermediate certificate."
          - "The identifier must be of the form
             V(A8:4A:6A:63:04:7D:DD:BA:E6:D1:39:B7:A6:45:65:EF:F3:A8:EC:A1)."
        type: str
      authority_key_identifier:
        description:
          - "Checks for the AuthorityKeyIdentifier extension. This is an identifier based
             on the private key of the issuer of the intermediate certificate."
          - "The identifier must be of the form
             V(C4:A7:B1:A4:7B:2C:71:FA:DB:E1:4B:90:75:FF:C4:15:60:85:89:10)."
        type: str
"""

EXAMPLES = r"""
---
### Example with HTTP-01 challenge ###

- name: Create a challenge for sample.com using a account key from a variable
  community.crypto.acme_certificate_order_create:
    account_key_content: "{{ account_private_key }}"
    csr: /etc/pki/cert/csr/sample.com.csr
  register: sample_com_challenge

# Alternative first step:
- name: Create a challenge for sample.com using a account key from Hashi Vault
  community.crypto.acme_certificate_order_create:
    account_key_content: >-
      {{ lookup('community.hashi_vault.hashi_vault', 'secret=secret/account_private_key:value') }}
    csr: /etc/pki/cert/csr/sample.com.csr
  register: sample_com_challenge

# Alternative first step:
- name: Create a challenge for sample.com using a account key file
  community.crypto.acme_certificate_order_create:
    account_key_src: /etc/pki/cert/private/account.key
    csr_content: "{{ lookup('file', '/etc/pki/cert/csr/sample.com.csr') }}"
  register: sample_com_challenge

# Perform the necessary steps to fulfill the challenge. For example:
#
# - name: Copy http-01 challenges
#   ansible.builtin.copy:
#     dest: /var/www/{{ item.identifier }}/{{ item.challenges['http-01'].resource }}
#     content: "{{ item.challenges['http-01'].resource_value }}"
#   loop: "{{ sample_com_challenge.challenge_data }}"
#   when: "'http-01' in item.challenges"

- name: Let the challenge be validated
  community.crypto.acme_certificate_order_validate:
    account_key_src: /etc/pki/cert/private/account.key
    order_uri: "{{ sample_com_challenge.order_uri }}"
    challenge: http-01

- name: Retrieve the cert and intermediate certificate
  community.crypto.acme_certificate_order_finalize:
    account_key_src: /etc/pki/cert/private/account.key
    csr: /etc/pki/cert/csr/sample.com.csr
    order_uri: "{{ sample_com_challenge.order_uri }}"
    cert_dest: /etc/httpd/ssl/sample.com.crt
    fullchain_dest: /etc/httpd/ssl/sample.com-fullchain.crt
    chain_dest: /etc/httpd/ssl/sample.com-intermediate.crt

---
### Example with DNS challenge against production ACME server ###

- name: Create a challenge for sample.com using a account key file.
  community.crypto.acme_certificate_order_create:
    acme_directory: https://acme-v01.api.letsencrypt.org/directory
    account_key_src: /etc/pki/cert/private/account.key
    csr: /etc/pki/cert/csr/sample.com.csr
  register: sample_com_challenge

# Perform the necessary steps to fulfill the challenge. For example:
#
# - name: Create DNS records for dns-01 challenges
#   community.aws.route53:
#     zone: sample.com
#     record: "{{ item.key }}"
#     type: TXT
#     ttl: 60
#     state: present
#     wait: true
#     # Note: item.value is a list of TXT entries, and route53
#     # requires every entry to be enclosed in quotes
#     value: "{{ item.value | map('community.dns.quote_txt', always_quote=true) | list }}"
#   loop: "{{ sample_com_challenge.challenge_data_dns | dict2items }}"

- name: Let the challenge be validated
  community.crypto.acme_certificate_order_validate:
    acme_directory: https://acme-v01.api.letsencrypt.org/directory
    account_key_src: /etc/pki/cert/private/account.key
    order_uri: "{{ sample_com_challenge.order_uri }}"
    challenge: dns-01

- name: Retrieve the cert and intermediate certificate
  community.crypto.acme_certificate_order_finalize:
    acme_directory: https://acme-v01.api.letsencrypt.org/directory
    account_key_src: /etc/pki/cert/private/account.key
    csr: /etc/pki/cert/csr/sample.com.csr
    order_uri: "{{ sample_com_challenge.order_uri }}"
    cert_dest: /etc/httpd/ssl/sample.com.crt
    fullchain_dest: /etc/httpd/ssl/sample.com-fullchain.crt
    chain_dest: /etc/httpd/ssl/sample.com-intermediate.crt
"""

RETURN = """
account_uri:
  description: ACME account URI.
  returned: success
  type: str
all_chains:
  description:
    - When O(retrieve_all_alternates=true), the module will query the ACME server for
      alternate chains. This return value will contain a list of all chains returned,
      the first entry being the main chain returned by the server.
    - See L(Section 7.4.2 of RFC8555,https://tools.ietf.org/html/rfc8555#section-7.4.2)
      for details.
  returned: success and O(retrieve_all_alternates=true)
  type: list
  elements: dict
  contains:
    cert:
      description:
        - The leaf certificate itself, in PEM format.
      type: str
      returned: always
    chain:
      description:
        - The certificate chain, excluding the root, as concatenated PEM certificates.
      type: str
      returned: always
    full_chain:
      description:
        - The certificate chain, excluding the root, but including the leaf certificate,
          as concatenated PEM certificates.
      type: str
      returned: always
selected_chain:
  description:
    - The selected certificate chain.
    - If O(select_chain) is not specified, this will be the main chain returned by the
      ACME server.
  returned: success
  type: dict
  contains:
    cert:
      description:
        - The leaf certificate itself, in PEM format.
      type: str
      returned: always
    chain:
      description:
        - The certificate chain, excluding the root, as concatenated PEM certificates.
      type: str
      returned: always
    full_chain:
      description:
        - The certificate chain, excluding the root, but including the leaf certificate,
          as concatenated PEM certificates.
      type: str
      returned: always
"""

import typing as t

from ansible_collections.community.crypto.plugins.module_utils._acme.acme import (
    create_backend,
    create_default_argspec,
)
from ansible_collections.community.crypto.plugins.module_utils._acme.certificate import (
    ACMECertificateClient,
)
from ansible_collections.community.crypto.plugins.module_utils._acme.errors import (
    ModuleFailException,
)

if t.TYPE_CHECKING:
    from ansible_collections.community.crypto.plugins.module_utils._acme.certificates import (  # pragma: no cover
        CertificateChain,
    )


def main() -> t.NoReturn:
    argument_spec = create_default_argspec(with_certificate=True)
    argument_spec.update_argspec(
        order_uri={"type": "str", "required": True},
        cert_dest={"type": "path"},
        fullchain_dest={"type": "path"},
        chain_dest={"type": "path"},
        deactivate_authzs={
            "type": "str",
            "default": "always",
            "choices": ["never", "always", "on_error", "on_success"],
        },
        retrieve_all_alternates={"type": "bool", "default": False},
        select_chain={
            "type": "list",
            "elements": "dict",
            "options": {
                "test_certificates": {
                    "type": "str",
                    "default": "all",
                    "choices": ["first", "last", "all"],
                },
                "issuer": {"type": "dict"},
                "subject": {"type": "dict"},
                "subject_key_identifier": {"type": "str"},
                "authority_key_identifier": {"type": "str"},
            },
        },
    )
    module = argument_spec.create_ansible_module()

    backend = create_backend(module, needs_acme_v2=False)

    try:
        client = ACMECertificateClient(module=module, backend=backend)
        select_chain_matcher = client.parse_select_chain(module.params["select_chain"])
        other = {}
        done = False
        order = None
        try:
            # Step 1: load order
            order = client.load_order()

            download_all_chains = (
                len(select_chain_matcher) > 0
                or module.params["retrieve_all_alternates"]
            )
            changed = False
            alternate_chains: list[CertificateChain] | None
            if order.status == "valid":
                # Step 2 and 3: download certificate(s) and chain(s)
                cert, alternate_chains = client.download_certificate(
                    order,
                    download_all_chains=download_all_chains,
                )
            else:
                client.check_that_authorizations_can_be_used(order)

                # Step 2: wait for authorizations to validate
                pending_authzs = client.collect_pending_authzs(order)
                client.wait_for_validation(pending_authzs)

                # Step 3: finalize order, wait, then download certificate(s) and chain(s)
                cert, alternate_chains = client.get_certificate(
                    order,
                    download_all_chains=download_all_chains,
                )
                changed = True

            # Step 4: pick chain, write certificates, and provide return values
            if alternate_chains is not None:
                # Prepare return value for all alternate chains
                if module.params["retrieve_all_alternates"]:
                    all_chains = [cert.to_json()]
                    for alt_chain in alternate_chains:
                        all_chains.append(alt_chain.to_json())
                    other["all_chains"] = all_chains

                # Try to select alternate chain depending on criteria
                if select_chain_matcher:
                    matching_chain = client.find_matching_chain(
                        chains=[cert] + alternate_chains,
                        select_chain_matcher=select_chain_matcher,
                    )
                    if matching_chain:
                        cert = matching_chain
                    else:
                        module.debug("Found no matching alternative chain")

            if client.write_cert_chain(
                cert=cert,
                cert_dest=module.params["cert_dest"],
                fullchain_dest=module.params["fullchain_dest"],
                chain_dest=module.params["chain_dest"],
            ):
                changed = True

            done = True
        finally:
            if (
                module.params["deactivate_authzs"] == "always"
                or (module.params["deactivate_authzs"] == "on_success" and done)
                or (module.params["deactivate_authzs"] == "on_error" and not done)
            ) and order:
                client.deactivate_authzs(order)
        module.exit_json(
            changed=changed,
            account_uri=client.client.account_uri,
            selected_chain=cert.to_json(),
            **other,
        )
    except ModuleFailException as e:
        e.do_fail(module=module)


if __name__ == "__main__":
    main()



#!/usr/bin/python
#
# Copyright (c) 2026 Zun Yang Chin, (@zunyangc)
#
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function
__metaclass__ = type


DOCUMENTATION = '''
---
module: azure_rm_storagesharedirectory_info
version_added: "3.19.0"
short_description: Get information about directories inside an Azure file share
description:
    - Get properties for a specific directory inside an Azure Storage file share, or list the
      directories at a given path (default the share root).
options:
    resource_group:
        description:
            - Name of the resource group that contains the storage account.
        required: true
        type: str
    account_name:
        description:
            - Name of the parent storage account hosting the file share.
        required: true
        type: str
    share_name:
        description:
            - Name of the existing file share.
        required: true
        type: str
    name:
        description:
            - Path of the directory to return properties for, for example C(foo/bar).
            - If omitted, the module lists directories under C(path) (defaults to the share root).
        type: str
        aliases:
            - path
    recursive:
        description:
            - When listing directories, walk the tree and return every directory under the start path.
            - Ignored when C(name) is specified.
        type: bool
        default: false

extends_documentation_fragment:
    - azure.azcollection.azure

author:
    - Zun Yang Chin (@zunyangc)
'''

EXAMPLES = '''
---
- name: Get a single directory's properties
  azure.azcollection.azure_rm_storagesharedirectory_info:
    resource_group: myResourceGroup
    account_name: myStorageAccount
    share_name: myshare
    name: parent/child

- name: List directories at the share root
  azure.azcollection.azure_rm_storagesharedirectory_info:
    resource_group: myResourceGroup
    account_name: myStorageAccount
    share_name: myshare

- name: List every directory in the share recursively
  azure.azcollection.azure_rm_storagesharedirectory_info:
    resource_group: myResourceGroup
    account_name: myStorageAccount
    share_name: myshare
    recursive: true
'''

RETURN = '''
storagesharedirectories:
    description:
        - List of directories matching the request. Always a list, even for a single C(name) lookup.
    returned: always
    type: list
    elements: dict
    contains:
        name:
            description:
                - Leaf name of the directory.
            returned: always
            type: str
            sample: child
        path:
            description:
                - Full path of the directory within the file share.
            returned: always
            type: str
            sample: parent/child
        share:
            description:
                - Name of the parent file share.
            returned: always
            type: str
            sample: myshare
        account:
            description:
                - Name of the parent storage account.
            returned: always
            type: str
            sample: myStorageAccount
        etag:
            description:
                - The directory entity tag.
            returned: when properties are available
            type: str
        last_modified:
            description:
                - Time the directory was last modified.
            returned: when properties are available
            type: str
        metadata:
            description:
                - Metadata associated with the directory.
            returned: when properties are available
            type: dict
'''

try:
    from azure.core.exceptions import ResourceNotFoundError
except ImportError:
    pass

from ansible_collections.azure.azcollection.plugins.module_utils.azure_rm_common import AzureRMModuleBase


class AzureRMStorageShareDirectoryInfo(AzureRMModuleBase):

    def __init__(self):

        self.module_arg_spec = dict(
            resource_group=dict(type='str', required=True),
            account_name=dict(type='str', required=True),
            share_name=dict(type='str', required=True),
            name=dict(type='str', aliases=['path']),
            recursive=dict(type='bool', default=False),
        )
        self.results = dict(
            changed=False,
            storagesharedirectories=[],
        )

        self.resource_group = None
        self.account_name = None
        self.share_name = None
        self.name = None
        self.recursive = False

        super(AzureRMStorageShareDirectoryInfo, self).__init__(
            self.module_arg_spec,
            supports_check_mode=True,
            supports_tags=False,
            facts_module=True,
        )

    def exec_module(self, **kwargs):
        for key in list(self.module_arg_spec.keys()):
            setattr(self, key, kwargs[key])

        if self.name is not None:
            self.name = self.name.strip('/').strip()

        if self.name:
            self.results['storagesharedirectories'] = self._get_one(self.name)
        else:
            self.results['storagesharedirectories'] = self._list('', recursive=self.recursive)

        return self.results

    def _get_one(self, path):
        client = self.get_share_directory_client(
            resource_group_name=self.resource_group,
            storage_account_name=self.account_name,
            share_name=self.share_name,
            directory_path=path,
        )
        try:
            props = client.get_directory_properties()
        except ResourceNotFoundError:
            return []
        except Exception as exc:
            self.fail("Error retrieving directory '{0}': {1}".format(path, str(exc)))
        return [self._props_to_dict(path, props)]

    def _list(self, base_path, recursive):
        share_client = self.get_share_client(
            resource_group_name=self.resource_group,
            storage_account_name=self.account_name,
            share_name=self.share_name,
        )
        if base_path:
            iterable = share_client.get_directory_client(base_path).list_directories_and_files()
        else:
            iterable = share_client.list_directories_and_files()

        results = []
        try:
            entries = list(iterable)
        except Exception as exc:
            self.fail("Error listing share '{0}': {1}".format(self.share_name, str(exc)))

        for entry in entries:
            if not entry.get('is_directory'):
                continue
            child_path = entry['name'] if not base_path else base_path + '/' + entry['name']
            child_client = self.get_share_directory_client(
                resource_group_name=self.resource_group,
                storage_account_name=self.account_name,
                share_name=self.share_name,
                directory_path=child_path,
            )
            try:
                props = child_client.get_directory_properties()
            except Exception:
                props = None
            results.append(self._props_to_dict(child_path, props))
            if recursive:
                results.extend(self._list(child_path, recursive=True))
        return results

    def _props_to_dict(self, path, props):
        leaf = path.split('/')[-1] if path else ''
        result = dict(
            name=leaf,
            path=path,
            share=self.share_name,
            account=self.account_name,
        )
        if props is not None:
            etag = getattr(props, 'etag', None)
            last_modified = getattr(props, 'last_modified', None)
            result.update(dict(
                etag=etag.replace('"', '') if isinstance(etag, str) else etag,
                last_modified=str(last_modified) if last_modified is not None else None,
                metadata=getattr(props, 'metadata', None) or {},
            ))
        return result


def main():
    AzureRMStorageShareDirectoryInfo()


if __name__ == '__main__':
    main()



# Copyright (c) 2023, Felix Fontein <felix@fontein.de>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

# Note that this plugin util is **PRIVATE** to the collection. It can have breaking changes at any time.
# Do not use this from other collections or standalone plugins/modules!

from __future__ import annotations

import typing as t
from collections.abc import Callable

from ansible.errors import AnsibleError
from ansible.module_utils.basic import missing_required_lib
from ansible.module_utils.common.text.converters import to_native

from ansible_collections.community.dns.plugins.module_utils._resolver import (
    ResolverError,
)

DNSPYTHON_IMPORTERROR: ImportError | None
try:
    import dns  # pylint: disable=unused-import
    import dns.exception  # pylint: disable=unused-import
    import dns.message  # pylint: disable=unused-import
    import dns.name  # pylint: disable=unused-import
    import dns.query  # pylint: disable=unused-import
    import dns.rcode  # pylint: disable=unused-import
    import dns.rdatatype  # pylint: disable=unused-import
    import dns.resolver  # pylint: disable=unused-import
    import dns.version
except ImportError as exc:
    DNSPYTHON_IMPORTERROR = exc
else:
    DNSPYTHON_IMPORTERROR = None  # type: ignore  # TODO

if t.TYPE_CHECKING:
    _T = t.TypeVar("_T")


def guarded_run(
    runner: Callable[[], _T],
    error_class: type[Exception] = AnsibleError,
    server: str | None = None,
) -> _T:
    suffix = f" for {server}" if server is not None else ""
    try:
        return runner()
    except ResolverError as e:
        raise error_class(f"Unexpected resolving error{suffix}: {to_native(e)}") from e
    except dns.exception.DNSException as e:
        raise error_class(f"Unexpected DNS error{suffix}: {to_native(e)}") from e


def assert_requirements_present(plugin_name: str, plugin_type: str) -> None:
    if DNSPYTHON_IMPORTERROR is not None:
        msg = f'The {plugin_name} {plugin_type} plugin is missing requirements: {missing_required_lib("dnspython >= 2.0.0")}'
        raise AnsibleError(msg) from DNSPYTHON_IMPORTERROR
    if dns.version.MAJOR < 2:
        msg = (
            f'The {plugin_name} {plugin_type} plugin is missing requirements: {missing_required_lib("dnspython >= 2.0.0")}.'
            f" Found version {dns.version.version}."
        )
        raise AnsibleError(msg)



#!/usr/bin/python
# Copyright: Ansible Project
# GNU General Public License v3.0+ (see COPYING or
# https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function


DOCUMENTATION = """
module: nxos_vsan
short_description: Configuration of vsan for Cisco NXOS MDS Switches.
description:
- Configuration of vsan for Cisco MDS NXOS.
version_added: 1.0.0
author:
- Suhas Bharadwaj (@srbharadwaj) (subharad@cisco.com)
notes:
- Tested against Cisco MDS NX-OS 8.4(1)
options:
  vsan:
    description:
    - List of vsan details to be added or removed
    type: list
    elements: dict
    suboptions:
      id:
        description:
        - Vsan id
        required: true
        type: int
      name:
        description:
        - Name of the vsan
        type: str
      suspend:
        description:
        - suspend the vsan if True
        type: bool
      remove:
        description:
        - Removes the vsan if True
        type: bool
      interface:
        description:
        - List of vsan's interfaces to be added
        type: list
        elements: str
"""

EXAMPLES = """
- name: Test that vsan module works
  cisco.nxos.nxos_vsan:
    vsan:
      - id: 922
        interface:
          - fc1/1
          - fc1/2
          - port-channel 1
        name: vsan-SAN-A
        remove: false
        suspend: false
      - id: 923
        interface:
          - fc1/11
          - fc1/21
          - port-channel 2
        name: vsan-SAN-B
        remove: false
        suspend: true
      - id: 1923
        name: vsan-SAN-Old
        remove: true
"""

RETURN = """
commands:
  description: commands sent to the device
  returned: always
  type: list
  sample:
    - terminal dont-ask
    - vsan database
    - vsan 922 interface fc1/40
    - vsan 922 interface port-channel 155
    - no terminal dont-ask
"""

import re

from ansible.module_utils.basic import AnsibleModule

from ansible_collections.cisco.nxos.plugins.module_utils.network.nxos.nxos import (
    load_config,
    run_commands,
)


__metaclass__ = type


class Vsan(object):
    def __init__(self, vsanid):
        self.vsanid = vsanid
        self.vsanname = None
        self.vsanstate = None
        self.vsanoperstate = None
        self.vsaninterfaces = []


class GetVsanInfoFromSwitch(object):
    """docstring for GetVsanInfoFromSwitch"""

    def __init__(self, module):
        self.module = module
        self.vsaninfo = {}
        self.processShowVsan()
        self.processShowVsanMembership()

    def execute_show_vsan_cmd(self):
        output = execute_show_command("show vsan", self.module)[0]
        return output

    def execute_show_vsan_mem_cmd(self):
        output = execute_show_command("show vsan membership", self.module)[0]
        return output

    def processShowVsan(self):
        patv = r"^vsan\s+(\d+)\s+information"
        patnamestate = "name:(.*)state:(.*)"
        patoperstate = "operational state:(.*)"

        output = self.execute_show_vsan_cmd().split("\n")
        for o in output:
            z = re.match(patv, o.strip())
            if z:
                v = z.group(1).strip()
                self.vsaninfo[v] = Vsan(v)

            z1 = re.match(patnamestate, o.strip())
            if z1:
                n = z1.group(1).strip()
                s = z1.group(2).strip()
                self.vsaninfo[v].vsanname = n
                self.vsaninfo[v].vsanstate = s

            z2 = re.match(patoperstate, o.strip())
            if z2:
                oper = z2.group(1).strip()
                self.vsaninfo[v].vsanoperstate = oper

        # 4094/4079 vsan is always present
        self.vsaninfo["4079"] = Vsan("4079")
        self.vsaninfo["4094"] = Vsan("4094")

    def processShowVsanMembership(self):
        patv = r"^vsan\s+(\d+).*"
        output = self.execute_show_vsan_mem_cmd().split("\n")
        memlist = []
        v = None
        for o in output:
            z = re.match(patv, o.strip())
            if z:
                if v is not None:
                    self.vsaninfo[v].vsaninterfaces = memlist
                    memlist = []
                v = z.group(1)
            if "interfaces" not in o:
                llist = o.strip().split()
                memlist = memlist + llist
        self.vsaninfo[v].vsaninterfaces = memlist

    def getVsanInfoObjects(self):
        return self.vsaninfo


def execute_show_command(command, module, command_type="cli_show"):
    output = "text"
    commands = [{"command": command, "output": output}]
    return run_commands(module, commands)


def flatten_list(command_lists):
    flat_command_list = []
    for command in command_lists:
        if isinstance(command, list):
            flat_command_list.extend(command)
        else:
            flat_command_list.append(command)
    return flat_command_list


def main():
    vsan_element_spec = dict(
        id=dict(required=True, type="int"),
        name=dict(type="str"),
        remove=dict(type="bool"),
        suspend=dict(type="bool"),
        interface=dict(type="list", elements="str"),
    )

    argument_spec = dict(
        vsan=dict(type="list", elements="dict", options=vsan_element_spec),
    )

    module = AnsibleModule(argument_spec=argument_spec, supports_check_mode=True)
    warnings = list()
    messages = list()
    commands_executed = list()
    result = {"changed": False}

    obj = GetVsanInfoFromSwitch(module)
    dictSwVsanObjs = obj.getVsanInfoObjects()

    commands = []
    vsan_list = module.params["vsan"]

    for eachvsan in vsan_list:
        vsanid = str(eachvsan["id"])
        vsanname = eachvsan["name"]
        vsanremove = eachvsan["remove"]
        vsansuspend = eachvsan["suspend"]
        vsaninterface_list = eachvsan["interface"]

        if int(vsanid) < 1 or int(vsanid) >= 4095:
            module.fail_json(
                msg=vsanid + " - This is an invalid vsan. Supported vsan range is 1-4094",
            )

        if vsanid in dictSwVsanObjs.keys():
            sw_vsanid = vsanid
            sw_vsanname = dictSwVsanObjs[vsanid].vsanname
            sw_vsanstate = dictSwVsanObjs[vsanid].vsanstate
            sw_vsaninterfaces = dictSwVsanObjs[vsanid].vsaninterfaces
        else:
            sw_vsanid = None
            sw_vsanname = None
            sw_vsanstate = None
            sw_vsaninterfaces = []

        if vsanremove:
            # Negative case:
            if vsanid == "4079" or vsanid == "4094":
                messages.append(
                    str(vsanid) + " is a reserved vsan, hence cannot be removed",
                )
                continue
            if vsanid == sw_vsanid:
                commands.append("no vsan " + str(vsanid))
                messages.append("deleting the vsan " + str(vsanid))
            else:
                messages.append(
                    "There is no vsan "
                    + str(vsanid)
                    + " present in the switch. Hence there is nothing to delete",
                )
            continue
        else:
            # Negative case:
            if vsanid == "4079" or vsanid == "4094":
                messages.append(
                    str(vsanid) + " is a reserved vsan, and always present on the switch",
                )
            else:
                if vsanid == sw_vsanid:
                    messages.append(
                        "There is already a vsan "
                        + str(vsanid)
                        + " present in the switch. Hence there is nothing to configure",
                    )
                else:
                    commands.append("vsan " + str(vsanid))
                    messages.append("creating vsan " + str(vsanid))

        if vsanname is not None:
            # Negative case:
            if vsanid == "4079" or vsanid == "4094":
                messages.append(
                    str(vsanid) + " is a reserved vsan, and cannot be renamed",
                )
            else:
                if vsanname == sw_vsanname:
                    messages.append(
                        "There is already a vsan "
                        + str(vsanid)
                        + " present in the switch, which has the name "
                        + vsanname
                        + " Hence there is nothing to configure",
                    )
                else:
                    commands.append("vsan " + str(vsanid) + " name " + vsanname)
                    messages.append(
                        "setting vsan name to " + vsanname + " for vsan " + str(vsanid),
                    )

        if vsansuspend:
            # Negative case:
            if vsanid == "4079" or vsanid == "4094":
                messages.append(
                    str(vsanid) + " is a reserved vsan, and cannot be suspended",
                )
            else:
                if sw_vsanstate == "suspended":
                    messages.append(
                        "There is already a vsan "
                        + str(vsanid)
                        + " present in the switch, which is in suspended state ",
                    )
                else:
                    commands.append("vsan " + str(vsanid) + " suspend")
                    messages.append("suspending the vsan " + str(vsanid))
        else:
            if sw_vsanstate == "active":
                messages.append(
                    "There is already a vsan "
                    + str(vsanid)
                    + " present in the switch, which is in active state ",
                )
            else:
                commands.append("no vsan " + str(vsanid) + " suspend")
                messages.append("no suspending the vsan " + str(vsanid))

        if vsaninterface_list is not None:
            for each_interface_name in vsaninterface_list:
                # For fcip,port-channel,vfc-port-channel need to remove the
                # extra space to compare
                temp = re.sub(" +", "", each_interface_name)
                if temp in sw_vsaninterfaces:
                    messages.append(
                        each_interface_name
                        + " is already present in the vsan "
                        + str(vsanid)
                        + " interface list",
                    )
                else:
                    commands.append(
                        "vsan " + str(vsanid) + " interface " + each_interface_name,
                    )
                    messages.append(
                        "adding interface " + each_interface_name + " to vsan " + str(vsanid),
                    )

    if len(commands) != 0:
        commands = ["terminal dont-ask"] + ["vsan database"] + commands + ["no terminal dont-ask"]

    cmds = flatten_list(commands)
    commands_executed = cmds

    if commands_executed:
        if module.check_mode:
            module.exit_json(
                changed=False,
                commands=commands_executed,
                msg="Check Mode: No cmds issued to the hosts",
            )
        else:
            result["changed"] = True
            load_config(module, commands_executed)

    result["messages"] = messages
    result["commands"] = commands_executed
    result["warnings"] = warnings
    module.exit_json(**result)


if __name__ == "__main__":
    main()



# Copyright (c) 2012, Michael DeHaan, <michael.dehaan@gmail.com>
# Copyright (c) 2017 Ansible Project
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

DOCUMENTATION = r"""
author: Unknown (!UNKNOWN)
name: log_plays
type: notification
short_description: Write playbook output to log file
description:
  - This callback writes playbook output to a file per host in the C(/var/log/ansible/hosts) directory.
requirements:
  - Whitelist in configuration
  - A writeable C(/var/log/ansible/hosts) directory by the user executing Ansible on the controller
options:
  log_folder:
    default: /var/log/ansible/hosts
    description: The folder where log files are created.
    type: str
    env:
      - name: ANSIBLE_LOG_FOLDER
    ini:
      - section: callback_log_plays
        key: log_folder
"""

import json
import os
import time
from collections.abc import MutableMapping

from ansible.module_utils.common.text.converters import to_bytes
from ansible.parsing.ajson import AnsibleJSONEncoder
from ansible.plugins.callback import CallbackBase
from ansible.utils.path import makedirs_safe

# NOTE: in Ansible 1.2 or later general logging is available without
# this plugin, just set ANSIBLE_LOG_PATH as an environment variable
# or log_path in the DEFAULTS section of your ansible configuration
# file.  This callback is an example of per hosts logging for those
# that want it.


class CallbackModule(CallbackBase):
    """
    logs playbook results, per host, in /var/log/ansible/hosts
    """

    CALLBACK_VERSION = 2.0
    CALLBACK_TYPE = "notification"
    CALLBACK_NAME = "community.general.log_plays"
    CALLBACK_NEEDS_WHITELIST = True

    TIME_FORMAT = "%b %d %Y %H:%M:%S"

    @staticmethod
    def _make_msg(now, playbook, task_name, task_action, category, data):
        return f"{now} - {playbook} - {task_name} - {task_action} - {category} - {data}\n\n"

    def __init__(self):
        super().__init__()

    def set_options(self, task_keys=None, var_options=None, direct=None):
        super().set_options(task_keys=task_keys, var_options=var_options, direct=direct)

        self.log_folder = self.get_option("log_folder")

        if not os.path.exists(self.log_folder):
            makedirs_safe(self.log_folder)

    def log(self, result, category):
        data = result._result
        if isinstance(data, MutableMapping):
            if "_ansible_verbose_override" in data:
                # avoid logging extraneous data
                data = "omitted"
            else:
                data = data.copy()
                invocation = data.pop("invocation", None)
                data = json.dumps(data, cls=AnsibleJSONEncoder)
                if invocation is not None:
                    data = f"{json.dumps(invocation)} => {data} "

        path = os.path.join(self.log_folder, result._host.get_name())
        now = time.strftime(self.TIME_FORMAT, time.localtime())

        msg = to_bytes(self._make_msg(now, self.playbook, result._task.name, result._task.action, category, data))
        with open(path, "ab") as fd:
            fd.write(msg)

    def v2_runner_on_failed(self, result, ignore_errors=False):
        self.log(result, "FAILED")

    def v2_runner_on_ok(self, result):
        self.log(result, "OK")

    def v2_runner_on_skipped(self, result):
        self.log(result, "SKIPPED")

    def v2_runner_on_unreachable(self, result):
        self.log(result, "UNREACHABLE")

    def v2_runner_on_async_failed(self, result):
        self.log(result, "ASYNC_FAILED")

    def v2_playbook_on_start(self, playbook):
        self.playbook = playbook._file_name

    def v2_playbook_on_import_for_host(self, result, imported_file):
        self.log(result, "IMPORTED", imported_file)

    def v2_playbook_on_not_import_for_host(self, result, missing_file):
        self.log(result, "NOTIMPORTED", missing_file)



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright: Ansible Project
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

DOCUMENTATION = r"""
---
module: lambda_layer
version_added: 5.5.0
short_description: Creates an AWS Lambda layer or deletes an AWS Lambda layer version
description:
  - This module allows the management of AWS Lambda functions aliases via the Ansible
  - Creates an Lambda layer from a ZIP archive.
    Each time you call this module with the same layer name, a new version is created.
  - Deletes a version of an Lambda layer.
author: "Aubin Bikouo (@abikouo)"
options:
  state:
    description:
    - Determines if an Lambda layer should be created, or deleted. When set to V(present), an Lambda layer version will be
      created. If set to V(absent), an existing Lambda layer version will be deleted.
    type: str
    default: present
    choices: [ absent, present ]
  name:
    description:
    - The name or Amazon Resource Name (ARN) of the Lambda layer.
    type: str
    required: true
    aliases:
    - layer_name
  description:
    description:
    - The description of the version.
    - Ignored when O(state=absent).
    - Mutually exclusive with O(version).
    type: str
  content:
    description:
    - The function layer archive.
    - Required when O(state=present).
    - Ignored when O(state=absent).
    - Mutually exclusive with o(version).
    type: dict
    suboptions:
      s3_bucket:
        description:
        - The Amazon S3 bucket of the layer archive.
        type: str
      s3_key:
        description:
        - The Amazon S3 key of the layer archive.
        type: str
      s3_object_version:
        description:
        - For versioned objects, the version of the layer archive object to use.
        type: str
      zip_file:
        description:
        - Path to the base64-encoded file of the layer archive.
        type: path
  compatible_runtimes:
    description:
    - A list of compatible function runtimes.
    - Ignored when O(state=absent).
    - Mutually exclusive with O(version).
    type: list
    elements: str
  license_info:
    description:
    - The layer's software license. It can be any of an SPDX license identifier,
      the URL of a license hosted on the internet or the full text of the license.
    - Ignored when O(state=absent).
    - Mutually exclusive with O(version).
    type: str
  compatible_architectures:
    description:
    - A list of compatible instruction set architectures. For example, x86_64.
    - Mutually exclusive with O(version).
    type: list
    elements: str
  version:
    description:
    - The version number of the layer to delete.
    - Set to V(-1) to delete all versions for the specified layer name.
    - Required when O(state=absent).
    - Ignored when O(state=present).
    - Mutually exclusive with O(description), O(content), O(compatible_runtimes),
      O(license_info), O(compatible_architectures).
    type: int
extends_documentation_fragment:
  - amazon.aws.common.modules
  - amazon.aws.region.modules
  - amazon.aws.boto3
"""

EXAMPLES = r"""
---
# Create a new Python library layer version from a zip archive located into a S3 bucket
- name: Create a new python library layer
  amazon.aws.lambda_layer:
    state: present
    name: sample-layer
    description: 'My Python layer'
    content:
      s3_bucket: 'lambda-layers-us-west-2-123456789012'
      s3_key: 'python_layer.zip'
    compatible_runtimes:
      - python3.6
      - python3.7
    license_info: MIT
    compatible_architectures:
      - x86_64

# Create a layer version from a zip in the local filesystem
- name: Create a new layer from a zip in the local filesystem
  amazon.aws.lambda_layer:
    state: present
    name: sample-layer
    description: 'My Python layer'
    content:
      zip_file: 'python_layer.zip'
    compatible_runtimes:
      - python3.6
      - python3.7
    license_info: MIT
    compatible_architectures:
      - x86_64

# Delete a layer version
- name: Delete a layer version
  amazon.aws.lambda_layer:
    state: absent
    name: sample-layer
    version: 2

# Delete all versions of test-layer
- name: Delete all versions
  amazon.aws.lambda_layer:
    state: absent
    name: test-layer
    version: -1
"""

RETURN = r"""
layer_version:
  description: Info about the layer version that was created or deleted.
  returned: always
  type: list
  elements: dict
  contains:
    content:
        description: Details about the layer version.
        returned: O(state=present)
        type: complex
        contains:
          location:
            description: A link to the layer archive in Amazon S3 that is valid for 10 minutes.
            returned: O(state=present)
            type: str
            sample: "https://awslambda-us-east-1-layers.s3.us-east-1.amazonaws.com/snapshots/123456789012/pylayer-9da91deffd3b4941b8baeeae5daeffe4"
          code_sha256:
            description: The SHA-256 hash of the layer archive.
            returned: O(state=present)
            type: str
            sample: "VLluleJZ3HTwDrdYolSMrS+8iPwEkcoXXaegjXf+dmc="
          code_size:
            description: The size of the layer archive in bytes.
            returned: O(state=present)
            type: int
            sample: 9473675
          signing_profile_version_arn:
            description: The Amazon Resource Name (ARN) for a signing profile version.
            returned: When a signing profile is defined
            type: str
          signing_job_arn:
            description: The Amazon Resource Name (ARN) of a signing job.
            returned: When a signing profile is defined
            type: str
    layer_arn:
        description: The ARN of the layer.
        returned: if the layer version exists or has been created
        type: str
        sample: "arn:aws:lambda:eu-west-2:123456789012:layer:pylayer"
    layer_version_arn:
        description: The ARN of the layer version.
        returned: if the layer version exists or has been created
        type: str
        sample: "arn:aws:lambda:eu-west-2:123456789012:layer:pylayer:2"
    description:
        description: The description of the version.
        returned: O(state=present)
        type: str
    compatible_runtimes:
        description: A list of compatible runtimes.
        returned: if it was defined for the layer version.
        type: list
        sample: ["python3.7"]
    created_date:
        description: The date that the layer version was created, in ISO-8601 format (YYYY-MM-DDThh:mm:ss.sTZD).
        returned: if the layer version exists or has been created
        type: str
        sample: "2022-09-28T14:27:35.866+0000"
    version:
        description: The version number.
        returned: if the layer version exists or has been created
        type: int
        sample: 1
    license_info:
        description: The layer's software license.
        returned: if it was defined for the layer version.
        type: str
        sample: "GPL-3.0-only"
    compatible_architectures:
        description: A list of compatible instruction set architectures.
        returned: if it was defined for the layer version.
        type: list
"""

try:
    import botocore
except ImportError:
    pass  # Handled by AnsibleAWSModule

from ansible.module_utils.common.dict_transformations import camel_dict_to_snake_dict

from ansible_collections.amazon.aws.plugins.module_utils.modules import AnsibleAWSModule
from ansible_collections.amazon.aws.plugins.module_utils.retries import AWSRetry


@AWSRetry.jittered_backoff()
def _list_layer_versions(client, **params):
    paginator = client.get_paginator("list_layer_versions")
    return paginator.paginate(**params).build_full_result()


class LambdaLayerFailure(Exception):
    def __init__(self, exc, msg):
        self.exc = exc
        self.msg = msg
        super().__init__(self)


def list_layer_versions(lambda_client, name):
    try:
        layer_versions = _list_layer_versions(lambda_client, LayerName=name)["LayerVersions"]
        return [camel_dict_to_snake_dict(layer) for layer in layer_versions]
    except (botocore.exceptions.ClientError, botocore.exceptions.BotoCoreError) as e:
        raise LambdaLayerFailure(e, f"Unable to list layer versions for name {name}")


def create_layer_version(lambda_client, params, check_mode=False):
    if check_mode:
        return {"msg": "Create operation skipped - running in check mode", "changed": True}

    opt = {"LayerName": params.get("name"), "Content": {}}
    keys = [
        ("description", "Description"),
        ("compatible_runtimes", "CompatibleRuntimes"),
        ("license_info", "LicenseInfo"),
        ("compatible_architectures", "CompatibleArchitectures"),
    ]
    for k, d in keys:
        if params.get(k) is not None:
            opt[d] = params.get(k)

    # Read zip file if any
    zip_file = params["content"].get("zip_file")
    if zip_file is not None:
        with open(zip_file, "rb") as zf:
            opt["Content"]["ZipFile"] = zf.read()
    else:
        opt["Content"]["S3Bucket"] = params["content"].get("s3_bucket")
        opt["Content"]["S3Key"] = params["content"].get("s3_key")
        if params["content"].get("s3_object_version") is not None:
            opt["Content"]["S3ObjectVersion"] = params["content"].get("s3_object_version")

    try:
        layer_version = lambda_client.publish_layer_version(**opt)
        layer_version.pop("ResponseMetadata", None)
        return {"changed": True, "layer_versions": [camel_dict_to_snake_dict(layer_version)]}
    except (botocore.exceptions.BotoCoreError, botocore.exceptions.ClientError) as e:
        raise LambdaLayerFailure(e, "Failed to publish a new layer version (check that you have required permissions).")


def delete_layer_version(lambda_client, params, check_mode=False):
    name = params.get("name")
    version = params.get("version")
    layer_versions = list_layer_versions(lambda_client, name)
    deleted_versions = []
    changed = False
    for layer in layer_versions:
        if version == -1 or layer["version"] == version:
            deleted_versions.append(layer)
            changed = True
            if not check_mode:
                try:
                    lambda_client.delete_layer_version(LayerName=name, VersionNumber=layer["version"])
                except (botocore.exceptions.BotoCoreError, botocore.exceptions.ClientError) as e:
                    raise LambdaLayerFailure(
                        e, f"Failed to delete layer version LayerName={name}, VersionNumber={version}."
                    )
    return {"changed": changed, "layer_versions": deleted_versions}


def execute_module(module, lambda_client):
    try:
        state = module.params.get("state")
        f_operation = create_layer_version
        if state == "absent":
            f_operation = delete_layer_version

        module.exit_json(**f_operation(lambda_client, module.params, module.check_mode))
    except LambdaLayerFailure as e:
        module.fail_json_aws(e.exc, msg=e.msg)


def main():
    argument_spec = dict(
        state=dict(type="str", choices=["present", "absent"], default="present"),
        name=dict(type="str", required=True, aliases=["layer_name"]),
        description=dict(type="str"),
        content=dict(
            type="dict",
            options=dict(
                s3_bucket=dict(type="str"),
                s3_key=dict(type="str", no_log=False),
                s3_object_version=dict(type="str"),
                zip_file=dict(type="path"),
            ),
            required_together=[["s3_bucket", "s3_key"]],
            required_one_of=[["s3_bucket", "zip_file"]],
            mutually_exclusive=[["s3_bucket", "zip_file"]],
        ),
        compatible_runtimes=dict(type="list", elements="str"),
        license_info=dict(type="str"),
        compatible_architectures=dict(type="list", elements="str"),
        version=dict(type="int"),
    )

    module = AnsibleAWSModule(
        argument_spec=argument_spec,
        required_if=[
            ("state", "present", ["content"]),
            ("state", "absent", ["version"]),
        ],
        mutually_exclusive=[
            ["version", "description"],
            ["version", "content"],
            ["version", "compatible_runtimes"],
            ["version", "license_info"],
            ["version", "compatible_architectures"],
        ],
        supports_check_mode=True,
    )

    lambda_client = module.client("lambda")
    execute_module(module, lambda_client)


if __name__ == "__main__":
    main()



#!/usr/bin/python
# -*- coding: utf-8 -*-

# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function
__metaclass__ = type

ANSIBLE_METADATA = {'metadata_version': '1.1',
                    'status': ['preview'],
                    'supported_by': 'community'}

DOCUMENTATION = r'''
---
module: intersight_ethernet_adapter_policy
short_description: Ethernet Adapter Policy configuration for Cisco Intersight
description:
  - Manages Ethernet Adapter Policy configuration on Cisco Intersight.
  - Configure virtual ethernet interface settings including advanced features like RDMA, QoS, and offload capabilities.
  - For more information see L(Cisco Intersight,https://intersight.com/apidocs/vnic/EthAdapterPolicy/get/).
extends_documentation_fragment: intersight
options:
  state:
    description:
      - If C(present), will verify the resource is present and will create if needed.
      - If C(absent), will verify the resource is absent and will delete if needed.
    type: str
    choices: [present, absent]
    default: present
  organization:
    description:
      - The name of the Organization this resource is assigned to.
      - Profiles, Policies, and Pools that are created within a Custom Organization are applicable only to devices in the same Organization.
    type: str
    default: default
  name:
    description:
      - The name assigned to the Ethernet Adapter Policy.
      - The name must be between 1 and 64 alphanumeric characters, allowing special characters :-_.
    type: str
    required: true
  description:
    description:
      - The user-defined description for the Ethernet Adapter Policy.
      - Description can contain letters(a-z, A-Z), numbers(0-9), hyphen(-), period(.), colon(:), or an underscore(_).
    type: str
    aliases: [descr]
  tags:
    description:
      - List of tags in Key:<user-defined key> Value:<user-defined value> format.
    type: list
    elements: dict
  vxlan_enabled:
    description:
      - Status of the Virtual Extensible LAN Protocol on the virtual ethernet interface.
    type: bool
    default: false
  nvgre_enabled:
    description:
      - Status of the Network Virtualization using Generic Routing Encapsulation on the virtual ethernet interface.
    type: bool
    default: false
  arfs_enabled:
    description:
      - Status of Accelerated Receive Flow Steering on the virtual ethernet interface.
    type: bool
    default: false
  ptp_enabled:
    description:
      - Status of Precision Time Protocol (PTP) on the virtual ethernet interface.
      - PTP can be enabled only on one vNIC on an adapter.
    type: bool
    default: false
  advanced_filter:
    description:
      - Enables advanced filtering on the interface.
    type: bool
    default: false
  interrupt_scaling:
    description:
      - Enables Interrupt Scaling on the interface.
    type: bool
    default: false
  geneve_enabled:
    description:
      - GENEVE offload protocol allows you to create logical networks that span physical network boundaries.
      - Cannot be enabled simultaneously with ARFS.
    type: bool
    default: false
  etherchannel_pinning_enabled:
    description:
      - Enables EtherChannel Pinning to combine multiple physical links between two network switches into a single logical link.
      - Transmit Queue Count should be at least 2 to enable ether channel pinning.
    type: bool
    default: false
  roce_enabled:
    description:
      - If enabled sets RDMA over Converged Ethernet (RoCE) on this virtual interface.
      - Cannot be enabled simultaneously with NVGRE.
    type: bool
    default: false
  roce_queue_pairs:
    description:
      - The number of queue pairs per adapter. Recommended value = integer power of 2.
      - Only applicable when roce_enabled is true.
    type: int
    default: 256
  roce_memory_regions:
    description:
      - The number of memory regions per adapter. Recommended value = integer power of 2.
      - Only applicable when roce_enabled is true.
    type: int
    default: 131072
  roce_resource_groups:
    description:
      - The number of resource groups per adapter. Recommended value = be an integer power of 2 greater than or equal to the number of CPU cores.
      - Only applicable when roce_enabled is true.
    type: int
    default: 2
  roce_version:
    description:
      - Configure RDMA over Converged Ethernet (RoCE) version on the virtual interface.
      - Only RoCEv1 is supported on Cisco VIC 13xx series adapters and only RoCEv2 is supported on Cisco VIC 14xx series adapters.
      - Only applicable when roce_enabled is true.
    type: int
    choices: [1, 2]
    default: 1
  roce_class_of_service:
    description:
      - The Class of Service for RoCE on this virtual interface.
      - Only applicable when roce_enabled is true.
    type: int
    choices: [1, 2, 4, 5, 6]
    default: 5
  interrupt_count:
    description:
      - The number of interrupt resources to allocate. Typical value is be equal to the number of completion queue resources.
    type: int
    default: 8
  interrupt_mode:
    description:
      - Preferred driver interrupt mode.
      - MSIx - Message Signaled Interrupts (MSI) with the optional extension (recommended).
      - MSI - MSI only.
      - INTx - PCI INTx interrupts.
    type: str
    choices: ['MSIx', 'MSI', 'INTx']
    default: 'MSIx'
  interrupt_coalescing_time:
    description:
      - The time to wait between interrupts or the idle period that must be encountered before an interrupt is sent.
      - To turn off interrupt coalescing, enter 0 (zero) in this field.
    type: int
    default: 125
  interrupt_coalescing_type:
    description:
      - Interrupt Coalescing Type.
      - MIN - The system waits for the time specified in the Coalescing Time field before sending another interrupt event.
      - IDLE - The system does not send an interrupt until there is a period of no activity
        lasting at least as long as the time specified in the Coalescing Time field.
    type: str
    choices: ['MIN', 'IDLE']
    default: 'MIN'
  rx_queue_count:
    description:
      - The number of receive queue resources to allocate.
    type: int
    default: 4
  rx_ring_size:
    description:
      - The number of descriptors in each receive queue.
    type: int
    default: 512
  tx_queue_count:
    description:
      - The number of transmit queue resources to allocate.
    type: int
    default: 1
  tx_ring_size:
    description:
      - The number of descriptors in each transmit queue.
    type: int
    default: 256
  completion_queue_count:
    description:
      - The number of completion queue resources to allocate.
      - In general, the number of completion queue resources to allocate is equal to the number of transmit queue resources
        plus the number of receive queue resources.
    type: int
    default: 5
  completion_ring_size:
    description:
      - The number of descriptors in each completion queue.
    type: int
    default: 1
  uplink_failback_timeout:
    description:
      - Uplink Failback Timeout in seconds when uplink failover is enabled for a vNIC.
      - After a vNIC has started using its secondary interface, this setting controls how long
        the primary interface must be available before the system resumes using the primary
        interface for the vNIC.
    type: int
    default: 5
  tcp_tx_checksum:
    description:
      - When enabled, the CPU sends all packets to the hardware so that the checksum can be calculated.
    type: bool
    default: true
  tcp_rx_checksum:
    description:
      - When enabled, the CPU sends all packet checksums to the hardware for validation.
    type: bool
    default: true
  tcp_large_send:
    description:
      - Enables the CPU to send large packets to the hardware for segmentation.
    type: bool
    default: true
  tcp_large_receive:
    description:
      - Enables the reassembly of segmented packets in hardware before sending them to the CPU.
    type: bool
    default: true
  rss_enabled:
    description:
      - Receive Side Scaling allows the incoming traffic to be spread across multiple CPU cores.
    type: bool
    default: true
  rss_ipv4_hash:
    description:
      - When enabled, the IPv4 address is used for traffic distribution.
    type: bool
    default: true
  rss_ipv6_hash:
    description:
      - When enabled, the IPv6 address is used for traffic distribution.
    type: bool
    default: false
  rss_ipv6_ext_hash:
    description:
      - When enabled, the IPv6 extensions are used for traffic distribution.
    type: bool
    default: true
  rss_tcp_ipv4_hash:
    description:
      - When enabled, both the IPv4 address and TCP port number are used for traffic distribution.
    type: bool
    default: true
  rss_tcp_ipv6_hash:
    description:
      - When enabled, both the IPv6 address and TCP port number are used for traffic distribution.
    type: bool
    default: false
  rss_tcp_ipv6_ext_hash:
    description:
      - When enabled, both the IPv6 extensions and TCP port number are used for traffic distribution.
    type: bool
    default: true
  rss_udp_ipv4_hash:
    description:
      - When enabled, both the IPv4 address and UDP port number are used for traffic distribution.
    type: bool
    default: false
  rss_udp_ipv6_hash:
    description:
      - When enabled, both the IPv6 address and UDP port number are used for traffic distribution.
    type: bool
    default: false
author:
  - Ron Gershburg (@rgershbu)
'''

EXAMPLES = r'''
- name: Create a basic Ethernet Adapter Policy
  cisco.intersight.intersight_ethernet_adapter_policy:
    api_private_key: "{{ api_private_key }}"
    api_key_id: "{{ api_key_id }}"
    name: "basic-ethernet-adapter-policy"
    description: "Basic Ethernet adapter policy with default settings"
    tags:
      - Key: "Environment"
        Value: "Production"

- name: Create an Ethernet Adapter Policy with VXLAN and advanced features
  cisco.intersight.intersight_ethernet_adapter_policy:
    api_private_key: "{{ api_private_key }}"
    api_key_id: "{{ api_key_id }}"
    name: "advanced-ethernet-adapter-policy"
    description: "Advanced Ethernet adapter policy with VXLAN and PTP"
    vxlan_enabled: true
    ptp_enabled: true
    advanced_filter: true
    interrupt_scaling: true
    interrupt_count: 16
    rx_queue_count: 8
    tx_queue_count: 4
    completion_queue_count: 12
    tags:
      - Key: "Environment"
        Value: "Production"
      - Key: "Feature"
        Value: "Advanced"

- name: Create an Ethernet Adapter Policy with RoCE enabled
  cisco.intersight.intersight_ethernet_adapter_policy:
    api_private_key: "{{ api_private_key }}"
    api_key_id: "{{ api_key_id }}"
    name: "roce-ethernet-adapter-policy"
    description: "Ethernet adapter policy with RoCE v2 enabled"
    roce_enabled: true
    roce_queue_pairs: 256
    roce_memory_regions: 131072
    roce_resource_groups: 2
    roce_version: 2
    roce_class_of_service: 5
    tags:
      - Key: "Protocol"
        Value: "RoCE"

- name: Create an Ethernet Adapter Policy with EtherChannel pinning
  cisco.intersight.intersight_ethernet_adapter_policy:
    api_private_key: "{{ api_private_key }}"
    api_key_id: "{{ api_key_id }}"
    name: "etherchannel-adapter-policy"
    description: "Ethernet adapter policy with EtherChannel pinning"
    etherchannel_pinning_enabled: true
    tx_queue_count: 2
    completion_queue_count: 6
    tags:
      - Key: "Feature"
        Value: "EtherChannel"

- name: Create an Ethernet Adapter Policy with custom RSS settings
  cisco.intersight.intersight_ethernet_adapter_policy:
    api_private_key: "{{ api_private_key }}"
    api_key_id: "{{ api_key_id }}"
    name: "custom-rss-adapter-policy"
    description: "Ethernet adapter policy with custom RSS configuration"
    rss_enabled: true
    rss_ipv4_hash: true
    rss_ipv6_hash: false
    rss_tcp_ipv4_hash: false
    rss_tcp_ipv6_hash: false
    rss_udp_ipv4_hash: true
    rss_udp_ipv6_hash: true
    tags:
      - Key: "Feature"
        Value: "CustomRSS"

- name: Delete an Ethernet Adapter Policy
  cisco.intersight.intersight_ethernet_adapter_policy:
    api_private_key: "{{ api_private_key }}"
    api_key_id: "{{ api_key_id }}"
    name: "old-ethernet-adapter-policy"
    state: absent
'''

RETURN = r'''
api_response:
  description: The API response output returned by the specified resource.
  returned: always
  type: dict
  sample:
    {
        "Moid": "6889f3e8f654d9f7013baae4",
        "ObjectType": "vnic.EthAdapterPolicy",
        "Name": "basic-ethernet-adapter-policy",
        "Description": "Basic Ethernet adapter policy with default settings",
        "AdvancedFilter": false,
        "ArfsSettings": {
            "Enabled": false
        },
        "VxlanSettings": {
            "Enabled": false
        },
        "NvgreSettings": {
            "Enabled": false
        },
        "PtpSettings": {
            "Enabled": false
        },
        "RoceSettings": {
            "Enabled": false
        },
        "InterruptSettings": {
            "Count": 8,
            "Mode": "MSIx",
            "CoalescingTime": 125,
            "CoalescingType": "MIN"
        },
        "RxQueueSettings": {
            "Count": 4,
            "RingSize": 512
        },
        "TxQueueSettings": {
            "Count": 1,
            "RingSize": 256
        },
        "CompletionQueueSettings": {
            "Count": 5,
            "RingSize": 1
        },
        "TcpOffloadSettings": {
            "TxChecksum": true,
            "RxChecksum": true,
            "LargeSend": true,
            "LargeReceive": true
        },
        "RssSettings": true,
        "RssHashSettings": {
            "Ipv4Hash": true,
            "Ipv6Hash": true,
            "Ipv6ExtHash": false,
            "TcpIpv4Hash": true,
            "TcpIpv6Hash": true,
            "TcpIpv6ExtHash": false,
            "UdpIpv4Hash": false,
            "UdpIpv6Hash": false
        }
    }
'''


from ansible.module_utils.basic import AnsibleModule
from ansible_collections.cisco.intersight.plugins.module_utils.intersight import IntersightModule, intersight_argument_spec


def validate_configuration(module):
    """
    Validate configuration according to Intersight constraints.
    """
    # Constraint 1: Cannot enable RDMA over Converged Ethernet and NVGRE simultaneously
    if module.params['roce_enabled'] and module.params['nvgre_enabled']:
        module.fail_json(
            msg="Cannot enable RDMA over Converged Ethernet and NVGRE simultaneously."
        )

    # Constraint 2: Cannot enable 'GENEVE Offload' and 'Accelerated Receive Flow Steering (ARFS)' simultaneously
    if module.params['geneve_enabled'] and module.params['arfs_enabled']:
        module.fail_json(
            msg="Cannot enable 'GENEVE Offload' and 'Accelerated Receive Flow Steering (ARFS)' simultaneously."
        )

    # Constraint 3: Cannot enable ether channel pinning with transmit queue count less than 2
    if module.params['etherchannel_pinning_enabled'] and module.params['tx_queue_count'] < 2:
        module.fail_json(
            msg="Cannot enable ether channel pinning with transmit queue count less than 2. "
                "Current tx_queue_count is {}. Set tx_queue_count to at least 2.".format(
                    module.params['tx_queue_count']
                )
        )

    # Validate ranges
    if not (1 <= module.params['interrupt_count'] <= 1024):
        module.fail_json(msg="interrupt_count must be between 1 and 1024")

    if not (0 <= module.params['interrupt_coalescing_time'] <= 65535):
        module.fail_json(msg="interrupt_coalescing_time must be between 0 and 65535")

    if not (1 <= module.params['rx_queue_count'] <= 1000):
        module.fail_json(msg="rx_queue_count must be between 1 and 1000")

    if not (64 <= module.params['rx_ring_size'] <= 16384):
        module.fail_json(msg="rx_ring_size must be between 64 and 16384")

    if not (1 <= module.params['tx_queue_count'] <= 1000):
        module.fail_json(msg="tx_queue_count must be between 1 and 1000")

    if not (64 <= module.params['tx_ring_size'] <= 16384):
        module.fail_json(msg="tx_ring_size must be between 64 and 16384")

    if not (1 <= module.params['completion_queue_count'] <= 2000):
        module.fail_json(msg="completion_queue_count must be between 1 and 2000")

    if not (1 <= module.params['completion_ring_size'] <= 256):
        module.fail_json(msg="completion_ring_size must be between 1 and 256")

    if not (0 <= module.params['uplink_failback_timeout'] <= 600):
        module.fail_json(msg="uplink_failback_timeout must be between 0 and 600")

    # RoCE specific validations
    if module.params['roce_enabled']:
        if not (1 <= module.params['roce_queue_pairs'] <= 8192):
            module.fail_json(msg="roce_queue_pairs must be between 1 and 8192")

        if not (1 <= module.params['roce_memory_regions'] <= 524288):
            module.fail_json(msg="roce_memory_regions must be between 1 and 524288")

        if not (1 <= module.params['roce_resource_groups'] <= 128):
            module.fail_json(msg="roce_resource_groups must be between 1 and 128")


def main():
    argument_spec = intersight_argument_spec.copy()
    argument_spec.update(
        state=dict(type='str', choices=['present', 'absent'], default='present'),
        organization=dict(type='str', default='default'),
        name=dict(type='str', required=True),
        description=dict(type='str', aliases=['descr']),
        tags=dict(type='list', elements='dict'),
        # VXLAN, NVGRE, ARFS, PTP settings
        vxlan_enabled=dict(type='bool', default=False),
        nvgre_enabled=dict(type='bool', default=False),
        arfs_enabled=dict(type='bool', default=False),
        ptp_enabled=dict(type='bool', default=False),
        # Advanced features
        advanced_filter=dict(type='bool', default=False),
        interrupt_scaling=dict(type='bool', default=False),
        geneve_enabled=dict(type='bool', default=False),
        etherchannel_pinning_enabled=dict(type='bool', default=False),
        # RoCE settings
        roce_enabled=dict(type='bool', default=False),
        roce_queue_pairs=dict(type='int', default=256),
        roce_memory_regions=dict(type='int', default=131072),
        roce_resource_groups=dict(type='int', default=2),
        roce_version=dict(type='int', choices=[1, 2], default=1),
        roce_class_of_service=dict(type='int', choices=[1, 2, 4, 5, 6], default=5),
        # Interrupt settings
        interrupt_count=dict(type='int', default=8),
        interrupt_mode=dict(type='str', choices=['MSIx', 'MSI', 'INTx'], default='MSIx'),
        interrupt_coalescing_time=dict(type='int', default=125),
        interrupt_coalescing_type=dict(type='str', choices=['MIN', 'IDLE'], default='MIN'),
        # Queue settings
        rx_queue_count=dict(type='int', default=4),
        rx_ring_size=dict(type='int', default=512),
        tx_queue_count=dict(type='int', default=1),
        tx_ring_size=dict(type='int', default=256),
        # Completion
        completion_queue_count=dict(type='int', default=5),
        completion_ring_size=dict(type='int', default=1),
        uplink_failback_timeout=dict(type='int', default=5),
        # TCP offload settings
        tcp_tx_checksum=dict(type='bool', default=True),
        tcp_rx_checksum=dict(type='bool', default=True),
        tcp_large_send=dict(type='bool', default=True),
        tcp_large_receive=dict(type='bool', default=True),
        # RSS settings
        rss_enabled=dict(type='bool', default=True),
        rss_ipv4_hash=dict(type='bool', default=True),
        rss_ipv6_hash=dict(type='bool', default=False),
        rss_ipv6_ext_hash=dict(type='bool', default=True),
        rss_tcp_ipv4_hash=dict(type='bool', default=True),
        rss_tcp_ipv6_hash=dict(type='bool', default=False),
        rss_tcp_ipv6_ext_hash=dict(type='bool', default=True),
        rss_udp_ipv4_hash=dict(type='bool', default=False),
        rss_udp_ipv6_hash=dict(type='bool', default=False),
    )

    module = AnsibleModule(
        argument_spec,
        supports_check_mode=True,
    )

    validate_configuration(module)
    intersight = IntersightModule(module)
    intersight.result['api_response'] = {}
    intersight.result['trace_id'] = ''

    # Resource path used to configure policy
    resource_path = '/vnic/EthAdapterPolicies'
    # Define API body used in compares or create
    intersight.api_body = {
        'Organization': {
            'Name': intersight.module.params['organization'],
        },
        'Name': intersight.module.params['name']
    }

    if module.params['state'] == 'present':
        intersight.set_tags_and_description()

        # Build the API body with all the settings
        intersight.api_body.update({
            'VxlanSettings': {
                'Enabled': module.params['vxlan_enabled']
            },
            'NvgreSettings': {
                'Enabled': module.params['nvgre_enabled']
            },
            'ArfsSettings': {
                'Enabled': module.params['arfs_enabled']
            },
            'PtpSettings': {
                'Enabled': module.params['ptp_enabled']
            },
            'AdvancedFilter': module.params['advanced_filter'],
            'InterruptScaling': module.params['interrupt_scaling'],
            'GeneveEnabled': module.params['geneve_enabled'],
            'EtherChannelPinningEnabled': module.params['etherchannel_pinning_enabled'],
            'RoceSettings': {
                'Enabled': module.params['roce_enabled']
            },
            'InterruptSettings': {
                'Count': module.params['interrupt_count'],
                'Mode': module.params['interrupt_mode'],
                'CoalescingTime': module.params['interrupt_coalescing_time'],
                'CoalescingType': module.params['interrupt_coalescing_type']
            },
            'RxQueueSettings': {
                'Count': module.params['rx_queue_count'],
                'RingSize': module.params['rx_ring_size']
            },
            'TxQueueSettings': {
                'Count': module.params['tx_queue_count'],
                'RingSize': module.params['tx_ring_size']
            },
            'CompletionQueueSettings': {
                'Count': module.params['completion_queue_count'],
                'RingSize': module.params['completion_ring_size']
            },
            'UplinkFailbackTimeout': module.params['uplink_failback_timeout'],
            'TcpOffloadSettings': {
                'TxChecksum': module.params['tcp_tx_checksum'],
                'RxChecksum': module.params['tcp_rx_checksum'],
                'LargeSend': module.params['tcp_large_send'],
                'LargeReceive': module.params['tcp_large_receive']
            },
            'RssSettings': module.params['rss_enabled'],
            'RssHashSettings': {
                'Ipv4Hash': module.params['rss_ipv4_hash'],
                'Ipv6Hash': module.params['rss_ipv6_hash'],
                'Ipv6ExtHash': module.params['rss_ipv6_ext_hash'],
                'TcpIpv4Hash': module.params['rss_tcp_ipv4_hash'],
                'TcpIpv6Hash': module.params['rss_tcp_ipv6_hash'],
                'TcpIpv6ExtHash': module.params['rss_tcp_ipv6_ext_hash'],
                'UdpIpv4Hash': module.params['rss_udp_ipv4_hash'],
                'UdpIpv6Hash': module.params['rss_udp_ipv6_hash']
            }
        })

        # Add RoCE-specific settings only when RoCE is enabled
        if module.params['roce_enabled']:
            intersight.api_body['RoceSettings'].update({
                'QueuePairs': module.params['roce_queue_pairs'],
                'MemoryRegions': module.params['roce_memory_regions'],
                'ResourceGroups': module.params['roce_resource_groups'],
                'Version': module.params['roce_version'],
                'ClassOfService': module.params['roce_class_of_service']
            })

    intersight.configure_policy_or_profile(resource_path=resource_path)

    module.exit_json(**intersight.result)


if __name__ == '__main__':
    main()



#!/usr/bin/python
# -*- coding: utf-8 -*-

# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function
__metaclass__ = type

ANSIBLE_METADATA = {'metadata_version': '1.1',
                    'status': ['preview'],
                    'supported_by': 'community'}

DOCUMENTATION = r'''
---
module: intersight_ldap_policy_info
short_description: Gather information about LDAP Policies in Cisco Intersight
description:
  - Gather information about LDAP Policies in L(Cisco Intersight,https://intersight.com).
  - Information can be filtered by O(organization) and O(name).
  - If no filters are passed, all LDAP Policies will be returned.
  - This module retrieves LDAP policy details including associated groups and providers.
extends_documentation_fragment: intersight
options:
  organization:
    description:
      - The name of the organization the LDAP Policy belongs to.
    type: str
  name:
    description:
      - The name of the LDAP Policy to gather information from.
    type: str
author:
  - Ron Gershburg (@rgershbu)
'''

EXAMPLES = r'''
- name: Fetch a specific LDAP Policy by name
  cisco.intersight.intersight_ldap_policy_info:
    api_private_key: "{{ api_private_key }}"
    api_key_id: "{{ api_key_id }}"
    name: "ldap-dns-policy"

- name: Fetch all LDAP Policies in a specific Organization
  cisco.intersight.intersight_ldap_policy_info:
    api_private_key: "{{ api_private_key }}"
    api_key_id: "{{ api_key_id }}"
    organization: "org_test"

- name: Fetch all LDAP Policies
  cisco.intersight.intersight_ldap_policy_info:
    api_private_key: "{{ api_private_key }}"
    api_key_id: "{{ api_key_id }}"
'''

RETURN = r'''
api_response:
  description: The API response output returned by the specified resource.
  returned: always
  type: dict
  sample:
    "api_response": [
    {
        "Name": "ldap-dns-policy",
        "ObjectType": "iam.LdapPolicy",
        "Enabled": true,
        "BaseProperties": {
            "BaseDn": "dc=example,dc=com",
            "Domain": "example.com",
            "Timeout": 30,
            "EnableEncryption": true,
            "BindMethod": "LoginCredentials",
            "Filter": "sAMAccountName",
            "GroupAttribute": "memberOf",
            "Attribute": "CiscoAvPair",
            "EnableGroupAuthorization": true,
            "EnableNestedGroupSearch": true,
            "NestedGroupSearchDepth": 64
        },
        "EnableDns": true,
        "DnsParameters": {
            "Source": "Configured",
            "SearchDomain": "example.com",
            "SearchForest": "example.com"
        },
        "UserSearchPrecedence": "LocalUserDb",
        "Groups": [
            {
                "Name": "admin-group",
                "ObjectType": "iam.LdapGroup",
                "GroupDn": "cn=admins,ou=groups,dc=example,dc=com",
                "Domain": "example.com",
                "Moid": "5dee1d736972652d321d26c5"
            },
            {
                "Name": "readonly-group",
                "ObjectType": "iam.LdapGroup",
                "GroupDn": "cn=readers,ou=groups,dc=example,dc=com",
                "Domain": "",
                "Moid": "5dee1d736972652d321d26c6"
            }
        ],
        "Providers": [],
        "Organization": {
            "Name": "default",
            "Moid": "5dee1d736972652d321d26b5",
            "ObjectType": "organization.Organization"
        },
        "Tags": [
            {
                "Key": "Environment",
                "Value": "Production"
            }
        ]
    },
    {
        "Name": "ldap-provider-policy",
        "ObjectType": "iam.LdapPolicy",
        "Enabled": true,
        "BaseProperties": {
            "BaseDn": "dc=company,dc=local",
            "Domain": "company.local",
            "Timeout": 0,
            "EnableEncryption": false,
            "BindMethod": "Anonymous",
            "Filter": "sAMAccountName",
            "GroupAttribute": "memberOf",
            "Attribute": "CiscoAvPair",
            "EnableGroupAuthorization": false,
            "EnableNestedGroupSearch": false
        },
        "EnableDns": false,
        "UserSearchPrecedence": "LocalUserDb",
        "Groups": [
            {
                "Name": "users-group",
                "ObjectType": "iam.LdapGroup",
                "GroupDn": "cn=users,ou=groups,dc=company,dc=local",
                "Domain": "",
                "Moid": "5dee1d736972652d321d26c7"
            }
        ],
        "Providers": [
            {
                "Server": "10.10.10.10",
                "Port": 389,
                "Vendor": "OpenLDAP",
                "ObjectType": "iam.LdapProvider",
                "Moid": "5dee1d736972652d321d26c8"
            },
            {
                "Server": "10.10.10.11",
                "Port": 389,
                "Vendor": "OpenLDAP",
                "ObjectType": "iam.LdapProvider",
                "Moid": "5dee1d736972652d321d26c9"
            }
        ],
        "Organization": {
            "Name": "default",
            "Moid": "5dee1d736972652d321d26b5",
            "ObjectType": "organization.Organization"
        },
        "Tags": []
    }
  ]
'''


from ansible.module_utils.basic import AnsibleModule
from ansible_collections.cisco.intersight.plugins.module_utils.intersight import IntersightModule, intersight_argument_spec


def main():
    argument_spec = intersight_argument_spec.copy()
    argument_spec.update(
        organization=dict(type='str'),
        name=dict(type='str')
    )
    module = AnsibleModule(
        argument_spec,
        supports_check_mode=True,
    )

    # Initialize the Intersight module
    intersight = IntersightModule(module)
    intersight.result['api_response'] = {}
    intersight.result['trace_id'] = ''

    # Resource path used to fetch info
    resource_path = '/iam/LdapPolicies'

    # Set query parameters
    query_params = intersight.set_query_params()

    # Get LDAP policies
    intersight.get_resource(
        resource_path=resource_path,
        query_params=query_params,
        return_list=True
    )

    # Fetch LDAP groups and providers for each LDAP policy
    ldap_policies = intersight.result['api_response']
    if isinstance(ldap_policies, list):
        for policy in ldap_policies:
            if policy.get('Moid'):
                # Fetch LDAP groups for this policy
                groups_query_params = {
                    '$filter': f"LdapPolicy.Moid eq '{policy['Moid']}'"
                }
                temp_intersight_groups = IntersightModule(module)
                temp_intersight_groups.get_resource(
                    resource_path='/iam/LdapGroups',
                    query_params=groups_query_params,
                    return_list=True
                )
                policy['Groups'] = temp_intersight_groups.result.get('api_response', [])

                # Fetch LDAP providers for this policy
                providers_query_params = {
                    '$filter': f"LdapPolicy.Moid eq '{policy['Moid']}'"
                }
                temp_intersight_providers = IntersightModule(module)
                temp_intersight_providers.get_resource(
                    resource_path='/iam/LdapProviders',
                    query_params=providers_query_params,
                    return_list=True
                )
                policy['Providers'] = temp_intersight_providers.result.get('api_response', [])

    elif isinstance(ldap_policies, dict) and ldap_policies.get('Moid'):
        # Single policy case
        groups_query_params = {
            '$filter': f"LdapPolicy.Moid eq '{ldap_policies['Moid']}'"
        }
        temp_intersight_groups = IntersightModule(module)
        temp_intersight_groups.get_resource(
            resource_path='/iam/LdapGroups',
            query_params=groups_query_params,
            return_list=True
        )
        ldap_policies['Groups'] = temp_intersight_groups.result.get('api_response', [])

        # Fetch LDAP providers for this policy
        providers_query_params = {
            '$filter': f"LdapPolicy.Moid eq '{ldap_policies['Moid']}'"
        }
        temp_intersight_providers = IntersightModule(module)
        temp_intersight_providers.get_resource(
            resource_path='/iam/LdapProviders',
            query_params=providers_query_params,
            return_list=True
        )
        ldap_policies['Providers'] = temp_intersight_providers.result.get('api_response', [])

    # Exit the module
    module.exit_json(**intersight.result)


if __name__ == '__main__':
    main()



#!/usr/bin/python

# Copyright (c) 2022, Gregory Furlong <gnfzdz@fzdz.io>
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

DOCUMENTATION = r"""
module: btrfs_subvolume
short_description: Manage btrfs subvolumes
version_added: "6.6.0"

description: Creates, updates and deletes btrfs subvolumes and snapshots.

options:
  automount:
    description:
      - Allow the module to temporarily mount the targeted btrfs filesystem in order to validate the current state and make
        any required changes.
    type: bool
    default: false
  default:
    description:
      - Make the subvolume specified by O(name) the filesystem's default subvolume.
    type: bool
    default: false
  filesystem_device:
    description:
      - A block device contained within the btrfs filesystem to be targeted.
      - Useful when multiple btrfs filesystems are present to specify which filesystem should be targeted.
    type: path
  filesystem_label:
    description:
      - A descriptive label assigned to the btrfs filesystem to be targeted.
      - Useful when multiple btrfs filesystems are present to specify which filesystem should be targeted.
    type: str
  filesystem_uuid:
    description:
      - A unique identifier assigned to the btrfs filesystem to be targeted.
      - Useful when multiple btrfs filesystems are present to specify which filesystem should be targeted.
    type: str
  name:
    description:
      - Name of the subvolume/snapshot to be targeted.
    required: true
    type: str
  recursive:
    description:
      - When true, indicates that parent/child subvolumes should be created/removedas necessary to complete the operation
        (for O(state=present) and O(state=absent) respectively).
    type: bool
    default: false
  snapshot_source:
    description:
      - Identifies the source subvolume for the created snapshot.
      - Infers that the created subvolume is a snapshot.
    type: str
  snapshot_conflict:
    description:
      - Policy defining behavior when a subvolume already exists at the path of the requested snapshot.
      - V(skip) - Create a snapshot only if a subvolume does not yet exist at the target location, otherwise indicate that
        no change is required. Warning, this option does not yet verify that the target subvolume was generated from a snapshot
        of the requested source.
      - V(clobber) - If a subvolume already exists at the requested location, delete it first. This option is not idempotent
        and results in a new snapshot being generated on every execution.
      - V(error) - If a subvolume already exists at the requested location, return an error. This option is not idempotent
        and results in an error on replay of the module.
    type: str
    choices: [skip, clobber, error]
    default: skip
  state:
    description:
      - Indicates the current state of the targeted subvolume.
    type: str
    choices: [absent, present]
    default: present

notes:
  - If any or all of the options O(filesystem_device), O(filesystem_label) or O(filesystem_uuid) parameters are provided,
    there is expected to be a matching btrfs filesystem. If none are provided and only a single btrfs filesystem exists or
    only a single btrfs filesystem is mounted, that filesystem is used; otherwise, the module takes no action and returns an
    error.
extends_documentation_fragment:
  - community.general._attributes

attributes:
  check_mode:
    support: partial
    details:
      - In some scenarios it may erroneously report intermediate subvolumes being created. After mounting, if a directory
        like file is found where the subvolume would have been created, the operation is skipped.
  diff_mode:
    support: none

author:
  - Gregory Furlong (@gnfzdz)
"""

EXAMPLES = r"""
- name: Create a @home subvolume under the root subvolume
  community.general.btrfs_subvolume:
    name: /@home
    filesystem_device: /dev/vda2

- name: Remove the @home subvolume if it exists
  community.general.btrfs_subvolume:
    name: /@home
    state: absent
    filesystem_device: /dev/vda2

- name: Create a snapshot of the root subvolume named @
  community.general.btrfs_subvolume:
    name: /@
    snapshot_source: /
    filesystem_device: /dev/vda2

- name: Create a snapshot of the root subvolume and make it the new default subvolume
  community.general.btrfs_subvolume:
    name: /@
    snapshot_source: /
    default: true
    filesystem_device: /dev/vda2

- name: Create a snapshot of the /@ subvolume and recursively creating intermediate subvolumes as required
  community.general.btrfs_subvolume:
    name: /@snapshots/@2022_06_09
    snapshot_source: /@
    recursive: true
    filesystem_device: /dev/vda2

- name: Remove the /@ subvolume and recursively delete child subvolumes as required
  community.general.btrfs_subvolume:
    name: /@snapshots/@2022_06_09
    snapshot_source: /@
    recursive: true
    filesystem_device: /dev/vda2
"""

RETURN = r"""
filesystem:
  description:
    - A summary of the final state of the targeted btrfs filesystem.
  type: dict
  returned: success
  contains:
    uuid:
      description: A unique identifier assigned to the filesystem.
      returned: success
      type: str
      sample: 96c9c605-1454-49b8-a63a-15e2584c208e
    label:
      description: An optional label assigned to the filesystem.
      returned: success
      type: str
      sample: Tank
    devices:
      description: A list of devices assigned to the filesystem.
      returned: success
      type: list
      sample:
        - /dev/sda1
        - /dev/sdb1
    default_subvolume:
      description: The ID of the filesystem's default subvolume.
      returned: success and if filesystem is mounted
      type: int
      sample: 5
    subvolumes:
      description: A list of dicts containing metadata for all of the filesystem's subvolumes.
      returned: success and if filesystem is mounted
      type: list
      elements: dict
      contains:
        id:
          description: An identifier assigned to the subvolume, unique within the containing filesystem.
          type: int
          sample: 256
        mountpoints:
          description: Paths where the subvolume is mounted on the targeted host.
          type: list
          sample: ["/home"]
        parent:
          description: The identifier of this subvolume's parent.
          type: int
          sample: 5
        path:
          description: The full path of the subvolume relative to the btrfs fileystem's root.
          type: str
          sample: /@home

modifications:
  description:
    - A list where each element describes a change made to the target btrfs filesystem.
  type: list
  returned: Success
  elements: str

target_subvolume_id:
  description:
    - The ID of the subvolume specified with the O(name) parameter, either pre-existing or created as part of module execution.
  type: int
  sample: 257
  returned: Success and subvolume exists after module execution
"""

import os
import tempfile

from ansible.module_utils.basic import AnsibleModule

from ansible_collections.community.general.plugins.module_utils._btrfs import (
    BtrfsCommands,
    BtrfsFilesystemsProvider,
    BtrfsModuleException,
    normalize_subvolume_path,
)


class BtrfsSubvolumeModule:
    __BTRFS_ROOT_SUBVOLUME = "/"
    __BTRFS_ROOT_SUBVOLUME_ID = 5
    __BTRFS_SUBVOLUME_INODE_NUMBER = 256

    __CREATE_SUBVOLUME_OPERATION = "create"
    __CREATE_SNAPSHOT_OPERATION = "snapshot"
    __DELETE_SUBVOLUME_OPERATION = "delete"
    __SET_DEFAULT_SUBVOLUME_OPERATION = "set-default"

    __UNKNOWN_SUBVOLUME_ID = "?"

    def __init__(self, module):
        self.module = module
        self.__btrfs_api = BtrfsCommands(module)
        self.__provider = BtrfsFilesystemsProvider(module)

        #  module parameters
        name = self.module.params["name"]
        self.__name = normalize_subvolume_path(name) if name is not None else None
        self.__state = self.module.params["state"]

        self.__automount = self.module.params["automount"]
        self.__default = self.module.params["default"]
        self.__filesystem_device = self.module.params["filesystem_device"]
        self.__filesystem_label = self.module.params["filesystem_label"]
        self.__filesystem_uuid = self.module.params["filesystem_uuid"]
        self.__recursive = self.module.params["recursive"]
        self.__snapshot_conflict = self.module.params["snapshot_conflict"]
        snapshot_source = self.module.params["snapshot_source"]
        self.__snapshot_source = normalize_subvolume_path(snapshot_source) if snapshot_source is not None else None

        # execution state
        self.__filesystem = None
        self.__required_mounts = []
        self.__unit_of_work = []
        self.__completed_work = []
        self.__temporary_mounts = dict()

    def run(self):
        error = None
        try:
            self.__load_filesystem()
            self.__prepare_unit_of_work()

            if not self.module.check_mode:
                # check required mounts & mount
                if len(self.__unit_of_work) > 0:
                    self.__execute_unit_of_work()
                    self.__filesystem.refresh()
            else:
                # check required mounts
                self.__completed_work.extend(self.__unit_of_work)
        except Exception as e:
            error = e
        finally:
            self.__cleanup_mounts()
            if self.__filesystem is not None:
                self.__filesystem.refresh_mountpoints()

        return (error, self.get_results())

    # Identify the targeted filesystem and obtain the current state
    def __load_filesystem(self):
        if self.__has_filesystem_criteria():
            filesystem = self.__find_matching_filesytem()
        else:
            filesystem = self.__find_default_filesystem()

        # The filesystem must be mounted to obtain the current state (subvolumes, default, etc)
        if not filesystem.is_mounted():
            if not self.__automount:
                raise BtrfsModuleException(
                    f"Target filesystem uuid={filesystem.uuid} is not currently mounted and automount=False."
                    "Mount explicitly before module execution or pass automount=True"
                )
            elif self.module.check_mode:
                # TODO is failing the module an appropriate outcome in this scenario?
                raise BtrfsModuleException(
                    f"Target filesystem uuid={filesystem.uuid} is not currently mounted. Unable to validate the current"
                    "state while running with check_mode=True"
                )
            else:
                self.__mount_subvolume_id_to_tempdir(filesystem, self.__BTRFS_ROOT_SUBVOLUME_ID)
                filesystem.refresh()
        self.__filesystem = filesystem

    def __has_filesystem_criteria(self):
        return (
            self.__filesystem_uuid is not None
            or self.__filesystem_label is not None
            or self.__filesystem_device is not None
        )

    def __find_matching_filesytem(self):
        criteria = {
            "uuid": self.__filesystem_uuid,
            "label": self.__filesystem_label,
            "device": self.__filesystem_device,
        }
        return self.__provider.get_matching_filesystem(criteria)

    def __find_default_filesystem(self):
        filesystems = self.__provider.get_filesystems()
        filesystem = None

        if len(filesystems) == 1:
            filesystem = filesystems[0]
        else:
            mounted_filesystems = [x for x in filesystems if x.is_mounted()]
            if len(mounted_filesystems) == 1:
                filesystem = mounted_filesystems[0]

        if filesystem is not None:
            return filesystem
        else:
            raise BtrfsModuleException(
                f"Failed to automatically identify targeted filesystem. No explicit device indicated and found {len(filesystems)} available filesystems."
            )

    # Prepare unit of work
    def __prepare_unit_of_work(self):
        if self.__state == "present":
            if self.__snapshot_source is None:
                self.__prepare_subvolume_present()
            else:
                self.__prepare_snapshot_present()

            if self.__default:
                self.__prepare_set_default()
        elif self.__state == "absent":
            self.__prepare_subvolume_absent()

    def __prepare_subvolume_present(self):
        subvolume = self.__filesystem.get_subvolume_by_name(self.__name)
        if subvolume is None:
            self.__prepare_before_create_subvolume(self.__name)
            self.__stage_create_subvolume(self.__name)

    def __prepare_before_create_subvolume(self, subvolume_name):
        closest_parent = self.__filesystem.get_nearest_subvolume(subvolume_name)
        self.__stage_required_mount(closest_parent)
        if self.__recursive:
            self.__prepare_create_intermediates(closest_parent, subvolume_name)

    def __prepare_create_intermediates(self, closest_subvolume, subvolume_name):
        relative_path = closest_subvolume.get_child_relative_path(self.__name)
        missing_subvolumes = [x for x in relative_path.split(os.path.sep) if len(x) > 0]
        if len(missing_subvolumes) > 1:
            current = closest_subvolume.path
            for s in missing_subvolumes[:-1]:
                separator = os.path.sep if current[-1] != os.path.sep else ""
                current = current + separator + s
                self.__stage_create_subvolume(current, True)

    def __prepare_snapshot_present(self):
        source_subvolume = self.__filesystem.get_subvolume_by_name(self.__snapshot_source)
        subvolume = self.__filesystem.get_subvolume_by_name(self.__name)
        subvolume_exists = subvolume is not None

        if subvolume_exists:
            if self.__snapshot_conflict == "skip":
                # No change required
                return
            elif self.__snapshot_conflict == "error":
                raise BtrfsModuleException(
                    f"Target subvolume={self.__name} already exists and snapshot_conflict='error'"
                )

        if source_subvolume is None:
            raise BtrfsModuleException(f"Source subvolume {self.__snapshot_source} does not exist")
        elif subvolume is not None and source_subvolume.id == subvolume.id:
            raise BtrfsModuleException("Snapshot source and target are the same.")
        else:
            self.__stage_required_mount(source_subvolume)

        if subvolume_exists and self.__snapshot_conflict == "clobber":
            self.__prepare_delete_subvolume_tree(subvolume)
        elif not subvolume_exists:
            self.__prepare_before_create_subvolume(self.__name)

        self.__stage_create_snapshot(source_subvolume, self.__name)

    def __prepare_subvolume_absent(self):
        subvolume = self.__filesystem.get_subvolume_by_name(self.__name)
        if subvolume is not None:
            self.__prepare_delete_subvolume_tree(subvolume)

    def __prepare_delete_subvolume_tree(self, subvolume):
        if subvolume.is_filesystem_root():
            raise BtrfsModuleException("Can not delete the filesystem's root subvolume")
        if not self.__recursive and len(subvolume.get_child_subvolumes()) > 0:
            raise BtrfsModuleException(
                f"Subvolume targeted for deletion {subvolume.path} has children and recursive=False."
                "Either explicitly delete the child subvolumes first or pass "
                "parameter recursive=True."
            )

        self.__stage_required_mount(subvolume.get_parent_subvolume())
        queue = self.__prepare_recursive_delete_order(subvolume) if self.__recursive else [subvolume]
        # prepare unit of work
        for s in queue:
            if s.is_mounted():
                # TODO potentially unmount the subvolume if automount=True ?
                raise BtrfsModuleException(f"Can not delete mounted subvolume={s.path}")
            if s.is_filesystem_default():
                self.__stage_set_default_subvolume(self.__BTRFS_ROOT_SUBVOLUME, self.__BTRFS_ROOT_SUBVOLUME_ID)
            self.__stage_delete_subvolume(s)

    def __prepare_recursive_delete_order(self, subvolume):
        """Return the subvolume and all descendents as a list, ordered so that descendents always occur before their ancestors"""
        pending = [subvolume]
        ordered = []
        while len(pending) > 0:
            next = pending.pop()
            ordered.append(next)
            pending.extend(next.get_child_subvolumes())
        ordered.reverse()  # reverse to ensure children are deleted before their parent
        return ordered

    def __prepare_set_default(self):
        subvolume = self.__filesystem.get_subvolume_by_name(self.__name)
        subvolume_id = subvolume.id if subvolume is not None else None

        if self.__filesystem.default_subvolid != subvolume_id:
            self.__stage_set_default_subvolume(self.__name, subvolume_id)

    # Stage operations to the unit of work
    def __stage_required_mount(self, subvolume):
        if subvolume.get_mounted_path() is None:
            if self.__automount:
                self.__required_mounts.append(subvolume)
            else:
                raise BtrfsModuleException(
                    f"The requested changes will require the subvolume '{subvolume.path}' to be mounted, but automount=False"
                )

    def __stage_create_subvolume(self, subvolume_path, intermediate=False):
        """
        Add required creation of an intermediate subvolume to the unit of work
        If intermediate is true, the action will be skipped if a directory like file is found at target
        after mounting a parent subvolume
        """
        self.__unit_of_work.append(
            {
                "action": self.__CREATE_SUBVOLUME_OPERATION,
                "target": subvolume_path,
                "intermediate": intermediate,
            }
        )

    def __stage_create_snapshot(self, source_subvolume, target_subvolume_path):
        """Add creation of a snapshot from source to target to the unit of work"""
        self.__unit_of_work.append(
            {
                "action": self.__CREATE_SNAPSHOT_OPERATION,
                "source": source_subvolume.path,
                "source_id": source_subvolume.id,
                "target": target_subvolume_path,
            }
        )

    def __stage_delete_subvolume(self, subvolume):
        """Add deletion of the target subvolume to the unit of work"""
        self.__unit_of_work.append(
            {
                "action": self.__DELETE_SUBVOLUME_OPERATION,
                "target": subvolume.path,
                "target_id": subvolume.id,
            }
        )

    def __stage_set_default_subvolume(self, subvolume_path, subvolume_id=None):
        """Add update of the filesystem's default subvolume to the unit of work"""
        self.__unit_of_work.append(
            {
                "action": self.__SET_DEFAULT_SUBVOLUME_OPERATION,
                "target": subvolume_path,
                "target_id": subvolume_id,
            }
        )

    # Execute the unit of work
    def __execute_unit_of_work(self):
        self.__check_required_mounts()
        for op in self.__unit_of_work:
            if op["action"] == self.__CREATE_SUBVOLUME_OPERATION:
                self.__execute_create_subvolume(op)
            elif op["action"] == self.__CREATE_SNAPSHOT_OPERATION:
                self.__execute_create_snapshot(op)
            elif op["action"] == self.__DELETE_SUBVOLUME_OPERATION:
                self.__execute_delete_subvolume(op)
            elif op["action"] == self.__SET_DEFAULT_SUBVOLUME_OPERATION:
                self.__execute_set_default_subvolume(op)
            else:
                raise ValueError(f"Unknown operation type '{op['action']}'")

    def __execute_create_subvolume(self, operation):
        target_mounted_path = self.__filesystem.get_mountpath_as_child(operation["target"])
        if not self.__is_existing_directory_like(target_mounted_path):
            self.__btrfs_api.subvolume_create(target_mounted_path)
            self.__completed_work.append(operation)

    def __execute_create_snapshot(self, operation):
        source_subvolume = self.__filesystem.get_subvolume_by_name(operation["source"])
        source_mounted_path = source_subvolume.get_mounted_path()
        target_mounted_path = self.__filesystem.get_mountpath_as_child(operation["target"])

        self.__btrfs_api.subvolume_snapshot(source_mounted_path, target_mounted_path)
        self.__completed_work.append(operation)

    def __execute_delete_subvolume(self, operation):
        target_mounted_path = self.__filesystem.get_mountpath_as_child(operation["target"])
        self.__btrfs_api.subvolume_delete(target_mounted_path)
        self.__completed_work.append(operation)

    def __execute_set_default_subvolume(self, operation):
        target = operation["target"]
        target_id = operation["target_id"]

        if target_id is None:
            target_subvolume = self.__filesystem.get_subvolume_by_name(target)

            if target_subvolume is None:
                self.__filesystem.refresh()  # the target may have been created earlier in module execution
                target_subvolume = self.__filesystem.get_subvolume_by_name(target)

            if target_subvolume is None:
                raise BtrfsModuleException(f"Failed to find existing subvolume '{target}'")
            else:
                target_id = target_subvolume.id

        self.__btrfs_api.subvolume_set_default(self.__filesystem.get_any_mountpoint(), target_id)
        self.__completed_work.append(operation)

    def __is_existing_directory_like(self, path):
        return os.path.exists(path) and (
            os.path.isdir(path) or os.stat(path).st_ino == self.__BTRFS_SUBVOLUME_INODE_NUMBER
        )

    def __check_required_mounts(self):
        filtered = self.__filter_child_subvolumes(self.__required_mounts)
        if len(filtered) > 0:
            for subvolume in filtered:
                self.__mount_subvolume_id_to_tempdir(self.__filesystem, subvolume.id)
            self.__filesystem.refresh_mountpoints()

    def __filter_child_subvolumes(self, subvolumes):
        """Filter the provided list of subvolumes to remove any that are a child of another item in the list"""
        filtered = []
        last = None
        ordered = sorted(subvolumes, key=lambda x: x.path)
        for next in ordered:
            if last is None or next.path[0 : len(last)] != last:
                filtered.append(next)
                last = next.path
        return filtered

    # Create/cleanup temporary mountpoints
    def __mount_subvolume_id_to_tempdir(self, filesystem, subvolid):
        # this check should be redundant
        if self.module.check_mode or not self.__automount:
            raise BtrfsModuleException(
                "Unable to temporarily mount required subvolumes"
                f" with automount={self.__automount} and check_mode={self.module.check_mode}"
            )

        cache_key = f"{filesystem.uuid}:{int(subvolid)}"
        # The subvolume was already mounted, so return the current path
        if cache_key in self.__temporary_mounts:
            return self.__temporary_mounts[cache_key]

        device = filesystem.devices[0]
        mountpoint = tempfile.mkdtemp(dir="/tmp")
        self.__temporary_mounts[cache_key] = mountpoint

        mount = self.module.get_bin_path("mount", required=True)
        command = [mount, "-o", f"noatime,subvolid={int(subvolid)}", device, mountpoint]
        self.module.run_command(command, check_rc=True)

        return mountpoint

    def __cleanup_mounts(self):
        for key in self.__temporary_mounts.keys():
            self.__cleanup_mount(self.__temporary_mounts[key])

    def __cleanup_mount(self, mountpoint):
        umount = self.module.get_bin_path("umount", required=True)
        result = self.module.run_command([umount, mountpoint])
        if result[0] == 0:
            rmdir = self.module.get_bin_path("rmdir", required=True)
            self.module.run_command([rmdir, mountpoint])

    # Format and return results
    def get_results(self):
        target = self.__filesystem.get_subvolume_by_name(self.__name)
        return dict(
            changed=len(self.__completed_work) > 0,
            filesystem=self.__filesystem.get_summary(),
            modifications=self.__get_formatted_modifications(),
            target_subvolume_id=(target.id if target is not None else None),
        )

    def __get_formatted_modifications(self):
        return [self.__format_operation_result(op) for op in self.__completed_work]

    def __format_operation_result(self, operation):
        action_type = operation["action"]
        if action_type == self.__CREATE_SUBVOLUME_OPERATION:
            return self.__format_create_subvolume_result(operation)
        elif action_type == self.__CREATE_SNAPSHOT_OPERATION:
            return self.__format_create_snapshot_result(operation)
        elif action_type == self.__DELETE_SUBVOLUME_OPERATION:
            return self.__format_delete_subvolume_result(operation)
        elif action_type == self.__SET_DEFAULT_SUBVOLUME_OPERATION:
            return self.__format_set_default_subvolume_result(operation)
        else:
            raise ValueError(f"Unknown operation type '{operation['action']}'")

    def __format_create_subvolume_result(self, operation):
        target = operation["target"]
        target_subvolume = self.__filesystem.get_subvolume_by_name(target)
        target_id = target_subvolume.id if target_subvolume is not None else self.__UNKNOWN_SUBVOLUME_ID
        return f"Created subvolume '{target}' ({target_id})"

    def __format_create_snapshot_result(self, operation):
        source = operation["source"]
        source_id = operation["source_id"]

        target = operation["target"]
        target_subvolume = self.__filesystem.get_subvolume_by_name(target)
        target_id = target_subvolume.id if target_subvolume is not None else self.__UNKNOWN_SUBVOLUME_ID
        return f"Created snapshot '{target}' ({target_id}) from '{source}' ({source_id})"

    def __format_delete_subvolume_result(self, operation):
        target = operation["target"]
        target_id = operation["target_id"]
        return f"Deleted subvolume '{target}' ({target_id})"

    def __format_set_default_subvolume_result(self, operation):
        target = operation["target"]
        if "target_id" in operation:
            target_id = operation["target_id"]
        else:
            target_subvolume = self.__filesystem.get_subvolume_by_name(target)
            target_id = target_subvolume.id if target_subvolume is not None else self.__UNKNOWN_SUBVOLUME_ID
        return f"Updated default subvolume to '{target}' ({target_id})"


def run_module():
    module_args = dict(
        automount=dict(type="bool", default=False),
        default=dict(type="bool", default=False),
        filesystem_device=dict(type="path"),
        filesystem_label=dict(type="str"),
        filesystem_uuid=dict(type="str"),
        name=dict(type="str", required=True),
        recursive=dict(type="bool", default=False),
        state=dict(type="str", default="present", choices=["present", "absent"]),
        snapshot_source=dict(type="str"),
        snapshot_conflict=dict(type="str", default="skip", choices=["skip", "clobber", "error"]),
    )

    module = AnsibleModule(argument_spec=module_args, supports_check_mode=True)

    subvolume = BtrfsSubvolumeModule(module)
    error, result = subvolume.run()
    if error is not None:
        module.fail_json(str(error), **result)
    else:
        module.exit_json(**result)


def main():
    run_module()


if __name__ == "__main__":
    main()



# Copyright 2016 Red Hat | Ansible
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

# Note that this module util is **PRIVATE** to the collection. It can have breaking changes at any time.
# Do not use this from other collections or standalone plugins/modules!

from __future__ import annotations

import base64
import random

from ansible.module_utils.common.text.converters import to_bytes, to_text


def generate_insecure_key() -> bytes:
    """Do NOT use this for cryptographic purposes!"""
    while True:
        # Generate a one-byte key. Right now the functions below do not use more
        # than one byte, so this is sufficient.
        key = bytes([random.randint(0, 255)])
        # Return anything that is not zero
        if key != b"\x00":
            return key


def scramble(value: str, key: bytes) -> str:
    """Do NOT use this for cryptographic purposes!"""
    if len(key) < 1:
        raise ValueError("Key must be at least one byte")
    b_value = to_bytes(value)
    k = key[0]
    b_value = bytes([k ^ b for b in b_value])
    return f"=S={to_text(base64.b64encode(b_value))}"


def unscramble(value: str, key: bytes) -> str:
    """Do NOT use this for cryptographic purposes!"""
    if len(key) < 1:
        raise ValueError("Key must be at least one byte")
    if not value.startswith("=S="):
        raise ValueError("Value does not start with indicator")
    b_value = base64.b64decode(value[3:])
    k = key[0]
    b_value = bytes([k ^ b for b in b_value])
    return to_text(b_value)



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright: (c) 2021, 2023, Anvitha Jain (@anvitha-jain) <anvjain@cisco.com>
# GNU General Public License v3.0+ (see LICENSE or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function

__metaclass__ = type

ANSIBLE_METADATA = {"metadata_version": "1.1", "status": ["preview"], "supported_by": "community"}

DOCUMENTATION = r"""
---
module: mso_schema_site_external_epg
short_description: Manage External EPG in schema of sites.
description:
- Manage External EPG in schema of sites on Cisco ACI Multi-Site.
- This module can only be used on versions of MSO that are 3.3 or greater.
author:
- Anvitha Jain (@anvitha-jain)
options:
  schema:
    description:
    - The name of the schema.
    type: str
    required: true
  template:
    description:
    - The name of the template to change.
    type: str
    required: true
  l3out:
    description:
    - The L3Out associated with the external epg.
    - Required when site is of type on-premise.
    - In NDO versions over 4.2, the parameter is accessible only when an external EPG is
    - linked to the current schema-template's VRF.
    type: str
    aliases: [ l3out_name ]
  l3out_schema:
    description:
    - The schema that defines the referenced L3Out.
    - If this parameter is unspecified, it defaults to the current schema.
    type: str
  l3out_template:
    description:
    - The template that defines the referenced L3Out.
    - If this parameter is unspecified, it defaults to the current template.
    type: str
  l3out_on_apic:
    description:
    - If this parameter is specified, the constructed l3out reference will refer to a distinguished name (DN) in APIC.
    type: bool
  external_epg:
    description:
    - The name of the External EPG to be managed.
    type: str
    aliases: [ name ]
  site:
    description:
    - The name of the site.
    type: str
    required: true
  route_reachability:
    description:
    - Configures if an external EPG route is pointing to the internet or to an external remote network.
    - Only available when associated with an azure site.
    type: str
    choices: [ internet, site-ext ]
    default: internet
  state:
    description:
    - Use C(present) or C(absent) for adding or removing.
    - Use C(query) for listing an object or multiple objects.
    type: str
    choices: [ absent, present, query ]
    default: present
seealso:
- module: cisco.mso.mso_schema_template_external_epg
extends_documentation_fragment: cisco.mso.modules
"""

EXAMPLES = r"""
- name: Add a Site External EPG
  cisco.mso.mso_schema_site_external_epg:
    host: mso_host
    username: admin
    password: SomeSecretPassword
    schema: Schema 1
    template: Template 1
    external_epg: External EPG 1
    l3out: L3out1
    state: present

- name: Remove a Site External EPG
  cisco.mso.mso_schema_site_external_epg:
    host: mso_host
    username: admin
    password: SomeSecretPassword
    schema: Schema 1
    template: Template 1
    external_epg: External EPG 1
    l3out: L3out1
    state: absent

- name: Query a Site External EPG
  cisco.mso.mso_schema_site_external_epg:
    host: mso_host
    username: admin
    password: SomeSecretPassword
    schema: Schema 1
    template: Template 1
    external_epg: External EPG 1
    l3out: L3out1
    state: query
  register: query_result

- name: Query all Site External EPGs
  cisco.mso.mso_schema_site_external_epg:
    host: mso_host
    username: admin
    password: SomeSecretPassword
    schema: Schema 1
    template: Template 1
    state: query
  register: query_result
"""

RETURN = r"""
"""

from ansible.module_utils.basic import AnsibleModule
from ansible_collections.cisco.mso.plugins.module_utils.mso import MSOModule, mso_argument_spec
from ansible_collections.cisco.mso.plugins.module_utils.schema import MSOSchema


def main():
    argument_spec = mso_argument_spec()
    argument_spec.update(
        schema=dict(type="str", required=True),
        template=dict(type="str", required=True),
        site=dict(type="str", required=True),
        l3out=dict(type="str", aliases=["l3out_name"]),
        l3out_schema=dict(type="str"),
        l3out_template=dict(type="str"),
        l3out_on_apic=dict(type="bool"),
        external_epg=dict(type="str", aliases=["name"]),
        route_reachability=dict(type="str", default="internet", choices=["internet", "site-ext"]),
        state=dict(type="str", default="present", choices=["absent", "present", "query"]),
    )

    module = AnsibleModule(
        argument_spec=argument_spec,
        supports_check_mode=True,
        required_if=[
            ["state", "absent", ["external_epg"]],
            ["state", "present", ["external_epg"]],
        ],
    )

    schema = module.params.get("schema")
    template = module.params.get("template").replace(" ", "")
    site = module.params.get("site")
    external_epg = module.params.get("external_epg")
    l3out = module.params.get("l3out")
    l3out_schema = module.params.get("l3out_schema")
    l3out_template = module.params.get("l3out_template")
    l3out_on_apic = module.params.get("l3out_on_apic")
    route_reachability = module.params.get("route_reachability")
    state = module.params.get("state")

    mso = MSOModule(module)

    l3out_template = template if l3out_template is None else l3out_template.replace(" ", "")
    l3out_schema = schema if l3out_schema is None else l3out_schema
    l3out_schema_id = mso.lookup_schema(l3out_schema)

    mso_schema = MSOSchema(mso, schema, template, site)
    mso_objects = mso_schema.schema_objects

    mso_schema.set_template_external_epg(external_epg, fail_module=False)

    # Path-based access uses site_id-template
    site_template = "{0}-{1}".format(mso_objects.get("site").details.get("siteId"), template)

    payload = dict()
    op_path = "/sites/{0}/externalEpgs/-".format(site_template)

    # Get template External EPG
    if mso_objects.get("template_external_epg") is not None:
        ext_epg_ref = mso_objects.get("template_external_epg").details.get("externalEpgRef")
        external_epgs = [e.get("externalEpgRef") for e in mso_objects.get("site").details.get("externalEpgs")]

        # Get Site External EPG
        if ext_epg_ref in external_epgs:
            external_epg_idx = external_epgs.index(ext_epg_ref)
            mso.existing = mso_objects.get("site").details.get("externalEpgs")[external_epg_idx]
            op_path = "/sites/{0}/externalEpgs/{1}".format(site_template, external_epg)

    ops = []
    l3out_dn = ""
    if state == "query":
        if external_epg is None:
            mso.existing = mso_objects.get("site").details.get("externalEpgs")
        elif not mso.existing:
            mso.fail_json(msg="External EPG '{external_epg}' not found".format(external_epg=external_epg))
        mso.exit_json()

    mso.previous = mso.existing

    if state == "absent":
        if mso.existing:
            mso.sent = mso.existing = {}
            ops.append(dict(op="remove", path=op_path))

    elif state == "present":
        # Get external EPGs type from template level and verify template_external_epg type.
        if mso_objects.get("template_external_epg") is not None and mso_objects.get("template_external_epg").details.get("extEpgType") != "cloud":
            if l3out is not None:
                path = "tenants/{0}".format(mso_objects.get("template").details.get("tenantId"))
                tenant_name = mso.request(path, method="GET").get("name")
                l3out_dn = "uni/tn-{0}/out-{1}".format(tenant_name, l3out)
            else:
                mso.fail_json(msg="L3Out cannot be empty when template external EPG type is 'on-premise'.")

        payload = dict(
            externalEpgRef=dict(
                schemaId=mso_schema.id,
                templateName=template,
                externalEpgName=external_epg,
            ),
            l3outDn=l3out_dn,
            routeReachabilityInternetType=route_reachability,
        )

        if not l3out_on_apic:
            payload.update(
                l3outRef=dict(
                    schemaId=l3out_schema_id,
                    templateName=l3out_template,
                    l3outName=l3out,
                ),
            )

        mso.sanitize(payload, collate=True)

        if mso.existing:
            ops.append(dict(op="replace", path=op_path, value=mso.sent))
        else:
            ops.append(dict(op="add", path=op_path, value=mso.sent))

        mso.existing = mso.proposed

    if not module.check_mode and mso.proposed != mso.previous:
        mso.request(mso_schema.path, method="PATCH", data=ops)

    mso.exit_json()


if __name__ == "__main__":
    main()



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright (c) 2021, Cisco Systems
# GNU General Public License v3.0+ (see LICENSE or
# https://www.gnu.org/licenses/gpl-3.0.txt)

DOCUMENTATION = r"""
module: organizations_wireless_clients_overview_by_device_info
short_description: Information module for organizations _wireless _clients _overview
  _by _device
description:
  - Get all organizations _wireless _clients _overview _by _device.
  - List access point client count at the moment in an organization.
version_added: '1.0.0'
extends_documentation_fragment:
  - cisco.meraki.module_info
  - cisco.meraki.module_info_pagination
author: Francisco Munoz (@fmunoz)
options:
  headers:
    description: Additional headers.
    type: dict
  organizationId:
    description:
      - OrganizationId path parameter. Organization ID.
    type: str
  networkIds:
    description:
      - >
        NetworkIds query parameter. Optional parameter to filter access points client
        counts by network ID. This filter uses multiple exact matches.
    elements: str
    type: list
  serials:
    description:
      - >
        Serials query parameter. Optional parameter to filter access points client
        counts by its serial numbers. This filter uses multiple exact matches.
    elements: str
    type: list
  campusGatewayClusterIds:
    description:
      - >
        CampusGatewayClusterIds query parameter. Optional parameter to filter access
        points client counts by MCG cluster IDs. This filter uses multiple exact matches.
    elements: str
    type: list
  perPage:
    description:
      - PerPage query parameter. The number of entries per page returned. Acceptable
        range is 3 - 1000. Default is 1000.
    type: int
  startingAfter:
    description:
      - >
        StartingAfter query parameter. A token used by the server to indicate the
        start of the page. Often this is a timestamp or an ID but it is not limited
        to those. This parameter should not be defined by client applications. The
        link for the first, last, prev, or next page in the HTTP Link header should
        define it.
    type: str
  endingBefore:
    description:
      - >
        EndingBefore query parameter. A token used by the server to indicate the end
        of the page. Often this is a timestamp or an ID but it is not limited to those.
        This parameter should not be defined by client applications. The link for
        the first, last, prev, or next page in the HTTP Link header should define
        it.
    type: str
requirements:
  - meraki >= 2.4.9
  - python >= 3.5
seealso:
  - name: Cisco Meraki documentation for wireless getOrganizationWirelessClientsOverviewByDevice
    description: Complete reference of the getOrganizationWirelessClientsOverviewByDevice
      API.
    link: https://developer.cisco.com/meraki/api-v1/#!get-organization-wireless-clients-overview-by-device
notes:
  - SDK Method used are
    wireless.Wireless.get_organization_wireless_clients_overview_by_device,
  - Paths used are
    get /organizations/{organizationId}/wireless/clients/overview/byDevice,
"""

EXAMPLES = r"""
- name: Get all organizations _wireless _clients _overview _by _device
  cisco.meraki.organizations_wireless_clients_overview_by_device_info:
    meraki_api_key: "{{ meraki_api_key }}"
    meraki_base_url: "{{ meraki_base_url }}"
    meraki_single_request_timeout: "{{ meraki_single_request_timeout }}"
    meraki_certificate_path: "{{ meraki_certificate_path }}"
    meraki_requests_proxy: "{{ meraki_requests_proxy }}"
    meraki_wait_on_rate_limit: "{{ meraki_wait_on_rate_limit }}"
    meraki_nginx_429_retry_wait_time: "{{ meraki_nginx_429_retry_wait_time }}"
    meraki_action_batch_retry_wait_time: "{{ meraki_action_batch_retry_wait_time }}"
    meraki_retry_4xx_error: "{{ meraki_retry_4xx_error }}"
    meraki_retry_4xx_error_wait_time: "{{ meraki_retry_4xx_error_wait_time }}"
    meraki_maximum_retries: "{{ meraki_maximum_retries }}"
    meraki_output_log: "{{ meraki_output_log }}"
    meraki_log_file_prefix: "{{ meraki_log_file_prefix }}"
    meraki_log_path: "{{ meraki_log_path }}"
    meraki_print_console: "{{ meraki_print_console }}"
    meraki_suppress_logging: "{{ meraki_suppress_logging }}"
    meraki_simulate: "{{ meraki_simulate }}"
    meraki_be_geo_id: "{{ meraki_be_geo_id }}"
    meraki_caller: "{{ meraki_caller }}"
    meraki_use_iterator_for_get_pages: "{{ meraki_use_iterator_for_get_pages }}"
    meraki_inherit_logging_config: "{{ meraki_inherit_logging_config }}"
    networkIds: []
    serials: []
    campusGatewayClusterIds: []
    perPage: 0
    startingAfter: string
    endingBefore: string
    organizationId: string
    total_pages: -1
    direction: next
  register: result
"""
RETURN = r"""
meraki_response:
  description: A dictionary or list with the response returned by the Cisco Meraki Python SDK
  returned: always
  type: dict
  sample: >
    {
      "items": [
        {
          "network": {
            "id": "string"
          },
          "serial": "string",
          "counts": {
            "byStatus": {
              "online": 0
            }
          }
        }
      ],
      "meta": {
        "counts": {
          "items": {
            "total": 0,
            "remaining": 0
          }
        }
      }
    }
"""
